import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.core.source import Source

from .. import utils

from .common import update_node_tick_type


class ReadParquet(Source):
    def __init__(
        self,
        parquet_file_path=None,
        where=None,
        time_assignment="end",
        discard_fields=None,
        fields=None,
        symbol_name_field=None,
        symbol=utils.adaptive,
        db=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        schema=None,
        **kwargs,
    ):
        """
        Read ticks from Parquet file

        Parameters
        ----------
        parquet_file_path: str
            Specifies the path or URL of the Parquet file to read.
        where: str, None
            Specifies a criterion for selecting the ticks to propagate.
        time_assignment: str
            Timestamps of the ticks created by `ReadParquet` are set to the start/end of the query or to the given
            tick field depending on the `time_assignment` parameter.
            Possible values are `start` and `end` (for `_START_TIME` and `_END_TIME`) or a field name.
            Default: `end`
        discard_fields: list, str, None
            A list of fields (`list` or comma-separated string) to be discarded from the output ticks.
        fields: list, str, None
            A list of fields (`list` or comma-separated string) to be picked from the output ticks.
            The opposite to `discard_fields`.
        symbol_name_field: str, None
            Field that is expected to contain the symbol name.
            When this parameter is set and one or more symbols containing time series
            (i.e. `[dbname]::[time series name]` and not just `[dbname]::`) are bound to the query or to this EP,
            only rows belonging to those symbols will be propagated.
        symbol: str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`
            Symbol(s) from which data should be taken.
        tick_type: str
            Tick type.
            Default: ANY.
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            Start time for tick generation. By default the start time of the query will be used.
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            End time for tick generation. By default the end time of the query will be used.
        schema: dict
            Dictionary of columns names with their types.
            You should set schema manually, if you want to use fields in `onetick-py` query description
            before its execution.
        kwargs:
            Deprecated. Use ``schema`` instead.
            Dictionary of columns names with their types.

        See also
        --------
        | **READ_FROM_PARQUET** OneTick event processor
        | :py:meth:`onetick.py.Source.write_parquet`

        Examples
        --------
        Simple Parquet file read:

        >>> data = otp.ReadParquet("/path/to/parquet/file")
        >>> otp.run(data)  # doctest: +SKIP

        Read Parquet file and filter fields:

        >>> data = otp.ReadParquet("/path/to/parquet/file", fields=["some_field", "another_field"])
        >>> otp.run(data)  # doctest: +SKIP

        Read Parquet file and filter rows:

        >>> data = otp.ReadParquet("/path/to/parquet/file", where="PRICE > 20")
        >>> otp.run(data)  # doctest: +SKIP
        """
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if parquet_file_path is None:
            raise ValueError("Missing required parameter `parquet_file_path`")

        if discard_fields and fields:
            raise ValueError("`discard_fields` and `fields` cannot be specified at the same time")

        if where is None:
            where = ""

        if discard_fields is None:
            discard_fields = ""
        elif isinstance(discard_fields, list):
            discard_fields = ",".join(discard_fields)

        if fields is None:
            fields = ""
        elif isinstance(fields, list):
            fields = ",".join(fields)

        if time_assignment in {"start", "end"}:
            time_assignment = f"_{time_assignment}_time".upper()

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(
                db=db,
                tick_type=tick_type,
                parquet_file_path=parquet_file_path,
                where=where,
                time_assignment=time_assignment,
                discard_fields=discard_fields,
                fields=fields,
                symbol_name_field=symbol_name_field,
            ),
            schema=schema,
            **kwargs,
        )

    def base_ep(
        self,
        parquet_file_path,
        where=None,
        time_assignment="end",
        discard_fields=None,
        fields=None,
        symbol_name_field=None,
        db=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
    ):
        if not hasattr(otq, "ReadFromParquet"):
            raise RuntimeError("Current version of OneTick don't support READ_FROM_PARQUET EP")

        node_kwargs = {}
        if symbol_name_field:
            node_kwargs["symbol_name_field"] = symbol_name_field

        src = Source(
            otq.ReadFromParquet(
                parquet_file_path=parquet_file_path,
                time_assignment=time_assignment,
                where=where,
                discard_fields=discard_fields,
                fields=fields,
                **node_kwargs,
            )
        )

        if db and tick_type:
            update_node_tick_type(src, tick_type, db)

        return src



import os
import argparse
import doctest
import pkgutil
import json
from collections import defaultdict
from typing import List, Optional
from jupyter_core.paths import jupyter_data_dir, jupyter_config_dir
import onetick.py as otp
from onetick.doc_utilities.ot_doctest import OTDoctestParser


def parse_args():
    parser = argparse.ArgumentParser(description='Adding snippets configuration from onetick-py package')
    parser.add_argument('extension',
                        choices=['snippets', 'snippets_menu', 'jupyterlab_snippets'],
                        default='snippets', nargs='?',
                        help='adding configuration to selected extension')
    parser.add_argument('-i', '--names-info', action='store_true',
                        help='show snippet names as yaml "tree"')
    args = parser.parse_args()
    return args


def collect_examples_from_obj(obj: doctest.DocTest) -> List:
    """groups examples for one object (docstring)"""
    snippets = []
    id_ = 1
    snippet = Snippet(obj_ref=f'{obj.name} {id_}')

    for example in obj.examples:
        snippet.append(example)
        if example.want:
            snippets.append(snippet)
            id_ += 1
            snippet = Snippet(obj_ref=f'{obj.name} {id_}')
    return snippets


def parse_string(string: str, caller: Optional[str] = None) -> str:
    parser = OTDoctestParser(caller=caller)
    lines = parser.parse(string)
    res: List = []
    snippet = Snippet()
    for item in lines:
        if isinstance(item, doctest.Example):
            snippet.append(item)
        else:
            if snippet.code:
                res.append(snippet)
                snippet = Snippet()
            res.append(item)
    return ''.join(map(str, res))


class SnippetInspectionError(Exception):
    pass


def inspect_modules():

    """Inspect onetick.py modules and collect snippets"""

    finder = doctest.DocTestFinder(parser=OTDoctestParser(caller='snippet'))
    snippets = Snippets()

    error_messages = []

    for _, modname, _ in pkgutil.walk_packages(otp.__path__, prefix=otp.__name__ + '.'):
        module = __import__(modname, fromlist=['__any__'])
        for doc in finder.find(module):
            for snippet in collect_examples_from_obj(doc):
                try:
                    snippets.append(snippet)
                except Exception as e:
                    error_messages.append(str(e))
    if error_messages:
        error_messages = '\t\n'.join(error_messages)
        message = f"During collection of snippets following exception were raised:\n{error_messages}"
        raise SnippetInspectionError(message)

    return snippets


def set_snippets(content):
    """Set nbextension-contrib-snippets with content"""
    snippets = {'snippets': content}
    path = jupyter_data_dir()
    path = os.path.join(path, 'nbextensions', 'snippets', 'snippets.json')
    if not os.path.exists(path):
        raise FileNotFoundError('Not able to set onetick snippets. Make sure nbextension snippets was installed.'
                                f'File {path} not found')
    with open(path, 'w') as f:
        json.dump(snippets, f, indent=4, sort_keys=True)


def set_snippets_menu(content):
    """Set nbextension-contrib-snippets_menu with content"""
    custom_js = f'''require(["nbextensions/snippets_menu/main"], function (snippets_menu) {{
    console.log('Loading `snippets_menu` customizations from `custom.js`');
    var horizontal_line = '---';
    var snippets = {{
        'name' : 'OneTick snippets',
        'sub-menu' : {json.dumps(content, indent=4, sort_keys=True)}
    }};
    snippets_menu.options['menus'].push(snippets);
    console.log('Loaded `snippets_menu` customizations from `custom.js`');
}});'''
    path = os.path.join(jupyter_config_dir(), 'custom', 'custom.js')
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w') as f:
        f.write(custom_js)


def set_jupyterlab_snippets(content):
    """Set jupyterlab-snippets with content"""

    def traverse_snippets(path, node_list):
        for node in node_list:
            if "sub-menu" in node:
                folder_path = os.path.join(path, node["name"])
                os.makedirs(folder_path, exist_ok=True)
                traverse_snippets(folder_path, node["sub-menu"])
            elif "snippet" in node:
                filename = os.path.join(path, node["name"].replace("|", " ").replace("/", " ")) + ".py"
                print(filename)
                with open(filename, 'w') as f:
                    f.write("\n".join(node["snippet"]))

    snippet_path = os.path.join(jupyter_data_dir(), "snippets", "OneTick snippets")
    os.makedirs(snippet_path, exist_ok=True)
    traverse_snippets(snippet_path, content)


class Snippet:

    """
    Snippet container.
    Contain code from doctest.Examples, add ability to set name.
    Validate names get from doctest.Examples
    """

    def __init__(self,
                 name: Optional[str] = None,
                 examples: Optional[List[doctest.Example]] = None,
                 obj_ref: Optional[str] = None):
        self._name = name
        self._examples = examples or []
        self._obj_ref = obj_ref
        self.skip_snippet = False

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if self._name:
            raise ValueError(f'Two names set for one snippet in "{self._obj_ref}": {self._name}, {value}')
        self._name = value

    @property
    def code(self) -> List[str]:
        res = []
        for ex in self._examples:
            res.extend(ex.source.strip().split('\n'))
        return res

    @property
    def code_raw(self) -> str:
        res = []
        for ex in self._examples:
            for i, line in enumerate(ex.source.strip().split('\n')):
                if i == 0:
                    res.append('>>> ' + line)
                else:
                    res.append('... ' + line)
            res.append(ex.want)
        return '\n'.join(res)

    def append(self, item: doctest.Example):
        for attr in ['name', 'skip_snippet']:
            if hasattr(item, attr):
                setattr(self, attr, getattr(item, attr))
        skip_example = False
        if hasattr(item, 'skip'):
            skip_example = getattr(item, 'skip')
        if not skip_example and item.source.strip():
            self._examples.append(item)

    def __str__(self):
        return self.code_raw


class Snippets:

    """Container for Snippet's"""

    def __init__(self):
        self._snippets = {}

    def append(self, item: Snippet):
        """
        Validate Snippet name;
        Add Snippet to container
        """
        if item.skip_snippet or not item.code:
            return
        if not item.name:
            return  # skipping nameless snippets
        if item.name in self._snippets:
            raise ValueError('Snippet name is not unique. '
                             f'Objects "{self._snippets[item.name]._obj_ref}", "{item._obj_ref}"')
        self._snippets[item.name] = item

    def extend(self, items: List):
        for item in items:
            self.append(item)

    def __iter__(self):
        yield from self._snippets.values()

    def dict_view(self):
        """Format snippets for nbextension-contrib-snippets"""
        return {"snippets": [{"name": s.name, "code": s.code} for s in self]}

    def menu_view(self):
        """Format snippets for nbextension-contrib-snippets_menu"""

        def tree():
            return defaultdict(tree)

        def set(t, keys: List, value):
            tmp = t
            for key in keys[:-1]:
                tmp = tmp[key]
            tmp[keys[-1]] = value

        def walk(t):
            res = []
            names = []
            for key in t.keys():
                if not isinstance(t[key], defaultdict):
                    res.append({"name": key, "snippet": t[key].code})
                    names.append(key)
                else:
                    code, name = walk(t[key])
                    res.append({"name": key, "sub-menu": code})
                    names.append({key: name})
            return res, names

        res = tree()

        for v in self:
            names = v.name.split('.')
            set(res, names, v)
        return walk(res)


def main():
    args = parse_args()
    snippets = inspect_modules()
    if args.names_info:
        import yaml
        _, menu = snippets.menu_view()
        print(yaml.dump(menu))
    else:
        if args.extension == 'snippets':
            set_snippets(snippets.dict_view())
        elif args.extension == 'snippets_menu':
            s, _ = snippets.menu_view()
            set_snippets_menu(s)
        elif args.extension == 'jupyterlab_snippets':
            s, _ = snippets.menu_view()
            set_jupyterlab_snippets(s)


if __name__ == '__main__':
    main()



import inspect
import types
import warnings

from onetick import py as otp
from onetick.py.core._source._symbol_param import _SymbolParamColumn, _SymbolParamSource
from onetick.py.core.column_operations.base import OnetickParameter


class _QueryEvalWrapper:
    def __init__(self, query, params=None, output_field=None, request_substitute_symbol=False,
                 generate_separate_file_only=False):
        self.query = query
        self.params = params
        self.output_field = output_field
        self.request_substitute_symbol = request_substitute_symbol
        self.generate_separate_file_only = generate_separate_file_only
        self._inner_source = None
        if isinstance(query, otp.Source):
            self._inner_source = query
            params = params or {}

            start, end = query._get_widest_time_range()
            if start and end and '_START_TIME' not in params and '_END_TIME' not in params:
                params['_START_TIME'] = start
                params['_END_TIME'] = end

            self.str_params = otp.query._params_to_str(params, with_expr=True)
        elif isinstance(query, otp.query):
            self.path = query.path
            self.str_params = query._params_to_str(params or {}, with_expr=True)
        else:
            raise ValueError("Wrong query parameter, it should be otp.query, otp.Query or function, "
                             "which returns them.")

    def _get_eval_string(self):
        eval_str = f'eval("{self.path}", "{self.str_params}")'
        if self.output_field:
            return f'{eval_str}.{self.output_field}'
        return eval_str

    def to_eval_string(self,
                       tmp_otq=None,
                       operation_suffix='eval',
                       start=None, end=None, timezone=None,
                       file_suffix='_eval_query.otq',
                       query_name='main_eval_query',
                       symbol_date=None) -> str:
        """
        If self._inner_source is not None, then temporary query needs to be saved
        or added to tmp_otq storage
        """
        if self._inner_source is not None:
            # if substitute symbol is requested, then we need to set an unbound symbol for query in eval
            # so that onetick can substitute it with the unbound symbol from the external query
            symbols = None
            if self.request_substitute_symbol:
                symbols = 'SYMBOL_TO_SUBSTITUTE'
            if not self.generate_separate_file_only and tmp_otq is not None:
                tmp_otq.merge(self._inner_source._tmp_otq)
                query_name = self._inner_source._store_in_tmp_otq(tmp_otq, symbols=symbols,
                                                                  start=start, end=end, timezone=timezone,
                                                                  operation_suffix=operation_suffix,
                                                                  symbol_date=symbol_date)
                self.path = f'THIS::{query_name}'
            else:
                self.path = self._inner_source.to_otq(file_suffix=file_suffix,
                                                      symbols=symbols,
                                                      query_name=query_name,
                                                      start=start, end=end, timezone=timezone,
                                                      symbol_date=symbol_date)
        return self._get_eval_string()

    def to_symbol_param(self):
        if self._inner_source:
            return self._inner_source.to_symbol_param()
        else:
            return _SymbolParamSource()

    def __str__(self):
        # TODO: deprecate after PY-952
        # warnings.warn('This method is deprecated, use to_eval_string() instead', FutureWarning)
        return self.to_eval_string()

    def copy(self, output_field=None):
        return _QueryEvalWrapper(query=self.query,
                                 params=self.params,
                                 output_field=output_field,
                                 request_substitute_symbol=self.request_substitute_symbol,
                                 generate_separate_file_only=self.generate_separate_file_only)

    def __getitem__(self, item):
        return self.copy(item)


def eval(query, symbol=None, start=None, end=None,
         generate_separate_file_only=False, **kwargs):
    """
    Creates an object with ``query`` with saved parameters that can be used later.

    It can be used to:

    * return a list of symbols for which the main query
      will be executed (multistage queries).
      Note that in this case ``query`` must return ticks with column **SYMBOL_NAME**.
    * return some value dynamically to be used in other places in the main query.
      Note that in this case ``query`` must return only single tick.

    Note that only constant expressions are allowed in query parameters,
    they must not depend on ticks.

    Parameters
    ----------
    query: :py:class:`onetick.py.Source`, :py:class:`onetick.py.query` or function
        source or query to evaluate.
        If function, then it must return :py:class:`onetick.py.Source` or :py:class:`onetick.py.query`.
        Parameter with name **symbol** and parameters specified in ``kwargs`` will be propagated
        to this function. Parameter from ``kwargs`` *must* be specified in function signature,
        but parameter **symbol** may be omitted if it is not used.
    symbol: :py:class:`~onetick.py.core._source._symbol_param._SymbolParamSource`
        symbol parameter that will be used by ``query`` as a symbol.
        If the function is used as a ``query``, parameter **symbol** can be defined in function
        signature and used in source operations.
    start: meta field (:py:class:`~onetick.py.core.source.MetaFields`) \
           or symbol param (:py:class:`~onetick.py.core._source._symbol_param._SymbolParamColumn`)
        start time with which ``query`` will be executed.
        By default the start time for evaluated query is inherited from the main query.
    end: meta field (:py:class:`~onetick.py.core.source.MetaFields`) \
         or symbol param (:py:class:`~onetick.py.core._source._symbol_param._SymbolParamColumn`)
        end time with which ``query`` will be executed.
        By default the end time for evaluated query is inherited from the main query.
    generate_separate_file_only: bool
        If set, sub-query will be generated in separate file.
        It's needed in some cases, e.g. when generating query for *otq_query_loader_daily.exe*,
        which executes all queries from a file.
    kwargs: str, int, meta fields (:py:class:`~onetick.py.core.source.MetaFields`) \
            or symbol params (:py:class:`~onetick.py.core._source._symbol_param._SymbolParamColumn`)
            or :py:meth:`~onetick.py.core.source.Source.join_with_query` parameters
        parameters that will be passed to ``query``.
        If the function is used as a ``query``, parameters specified in ``kwargs``
        *must* be defined in function signature and can be used in source operations.

    See also
    --------
    :ref:`api/misc/symbol_param:Symbol Parameters Objects`
    :ref:`static/concepts/symbols:Symbol parameters`

    Examples
    --------

    Use ``otp.eval`` to be passed as symbols when running the query:

    >>> def fsq():
    ...     symbols = otp.Ticks(SYMBOL_NAME=['AAPL', 'AAP'])
    ...     return symbols
    >>> main = otp.DataSource(db='US_COMP', tick_type='TRD', date=otp.dt(2022, 3, 1))
    >>> main['SYMBOL_NAME'] = main.Symbol.name
    >>> main = otp.merge([main], symbols=otp.eval(fsq))
    >>> otp.run(main)  # OTdirective: snippet-name: eval with symbols;
                         Time  PRICE  SIZE SYMBOL_NAME
    0 2022-03-01 00:00:00.000   1.30   100        AAPL
    1 2022-03-01 00:00:00.000  45.37     0         AAP
    2 2022-03-01 00:00:00.001   1.40    10        AAPL
    3 2022-03-01 00:00:00.001  45.41     0         AAP
    4 2022-03-01 00:00:00.002   1.40    50        AAPL

    Use ``otp.eval`` as filter:

    >>> def get_filter(a, b):
    ...     return otp.Tick(WHERE=f'X >= {str(a)} and X < {str(b)}', OTHER_FIELD='X')
    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> # note that in this case column WHERE must be specified,
    >>> # because evaluated query returns tick with more than one field
    >>> data, _ = data[otp.eval(get_filter, a=0, b=2)['WHERE']]
    >>> otp.run(data)
            Time  X
    0 2003-12-01  1

    Use ``otp.eval`` with meta fields:

    >>> def filter_by_tt(tick_type):
    ...     res = otp.Ticks({
    ...         'TICK_TYPE': ['TRD', 'QTE'],
    ...         'WHERE': ['PRICE>=1.4', 'ASK_PRICE>=1.4']
    ...     })
    ...     res, _ = res[res['TICK_TYPE'] == tick_type]
    ...     return res.drop(['TICK_TYPE'])
    >>> t = otp.DataSource('US_COMP::TRD')
    >>> # note that in this case column WHERE do not need to be specified,
    >>> # because evaluated query returns tick with only one field
    >>> t, _ = t[otp.eval(filter_by_tt, tick_type=t['_TICK_TYPE'])]
    >>> otp.run(t, start=otp.dt(2022, 3, 1), end=otp.dt(2022, 3, 2))
                         Time  PRICE  SIZE
    0 2022-03-01 00:00:00.001    1.4    10
    1 2022-03-01 00:00:00.002    1.4    50
    """
    if isinstance(query, (types.FunctionType, types.LambdaType)) or inspect.ismethod(query):
        params = {}
        params_to_convert = {}
        sig = inspect.signature(query)
        for param in sig.parameters:
            if "symbol" == param:
                if isinstance(symbol, _SymbolParamSource):
                    params['symbol'] = symbol
                else:
                    params["symbol"] = _SymbolParamSource()
            else:
                value = kwargs[param]
                if isinstance(value, otp.Column) and (value.name not in otp.Source.meta_fields
                                                      and not isinstance(value, _SymbolParamColumn)
                                                      and not isinstance(value, OnetickParameter)):
                    raise ValueError('Eval parameters can not depend on tick.')
                params_to_convert[param] = value
        params.update(prepare_params(**params_to_convert))

        query = query(**params)
    params = {}
    request_substitute_symbol = False
    if symbol is not None:
        if not isinstance(symbol, _SymbolParamSource):
            raise ValueError("Symbol parameter has wrong type, are you sure you are using it from function passed "
                             "to merge or join method?")
        params["SYMBOL_NAME"] = symbol.name
        request_substitute_symbol = True

    if start is not None:
        params["_START_TIME"] = start
    if end is not None:
        params["_END_TIME"] = end
    params.update(kwargs)

    return _QueryEvalWrapper(query, params, request_substitute_symbol=request_substitute_symbol,
                             generate_separate_file_only=generate_separate_file_only)


def prepare_params(**kwargs):
    converted_params = {}
    for key, value in kwargs.items():
        dtype = otp.types.get_object_type(value)
        # pylint: disable-next=unidiomatic-typecheck
        if type(value) is str and len(value) > otp.string.DEFAULT_LENGTH:
            dtype = otp.string[len(value)]
        param = OnetickParameter(key, dtype)
        converted_params[key] = param
    return converted_params



from typing import Optional

from onetick.py.otq import otq

from onetick.py.core.source import Source
from onetick.py.backports import Literal

from .. import utils
from ..compatibility import (
    is_data_file_query_supported,
    is_data_file_query_symbology_supported,
)

from .common import update_node_tick_type


class DataFile(Source):

    _PROPERTIES = Source._PROPERTIES + [
        '_file',
        '_msg_format',
        '_symbol_name_field',
        '_symbology',
        '_timestamp_column',
        '_time_assignment',
        '_file_contents',
        '_format_file',
        '_config_dir',
        '_db',
        '_tick_type',
    ]

    def __init__(
        self,
        file: Optional[str] = None,
        msg_format: Literal['arrow', 'json'] = 'arrow',
        symbol_name_field: str = 'SYMBOL_NAME',
        symbology: Optional[str] = None,
        timestamp_column: Optional[str] = None,
        time_assignment=utils.adaptive,
        file_contents: Optional[str] = None,
        format_file: Optional[str] = None,
        config_dir: Optional[str] = None,
        start=utils.adaptive,
        end=utils.adaptive,
        db=utils.adaptive,
        tick_type=utils.adaptive,
        symbols=utils.adaptive_to_default,
        schema=None,
        **kwargs,
    ):
        """
        Reads data streams in supported formats from file or file content,
        processes these streams to generate ticks,
        and queries the result against a list of symbols.

        Continuous event processing (CEP) of the data file is currently supported only for the JSON format.
        For other formats, or if the file content is specified,
        EP will function without continuous processing (no exception will be thrown).

        Note
        ----
        The method is supported only on 64-bit Windows/Linux platforms.

        Parameters
        ----------
        file:
            Specifies the path of the file to process.
            The parameter **DATA_FILE_PATH** in the OneTick configuration file specifies the set of directories
            where this file is searched for if the value of this parameter is a relative path.

            When run on a tick server with the **TICK_SERVER_DATA_CACHE_DIR** OneTick configuration variable
            pointing to a directory,
            this method will attempt to fetch the file from the host
            if the file is not found locally.
            Fetched files will be cached in this directory.

            To use AWS S3 file system paths,
            the **AWS_ACCESS_KEY_ID** and **AWS_SECRET_ACCESS_KEY** environment variables must be set.
            Note: Currently, AWS S3 is supported only for the ``arrow`` format.
        msg_format:
            Specifies the format of input data: *arrow* or *json*.
        symbol_name_field:
            Defines the field expected to contain the symbol name.
            Its data type should be either `string` or `varstring`.
            When one or more symbols are specified for the query or this method,
            only ticks corresponding to those symbols will be propagated.
        symbology:
            Defines the symbology of symbol names (values of ``symbol_name_field``) in the data file.
            If specified, to find correct time series in the file for the query symbol
            this method will first search for synonym[s] of the symbols
            in the specified symbology using reference database.
            Then it will find the history of this new symbol for the query interval
            and finally will query the data file using the symbol names from history.
        timestamp_column:
            If provided, the value of this field will be used to set the timestamps for the ticks.
            In the case of ``msg_format=arrow``, the Arrow type of this field must be `DATE64`, `TIMESTAMP`
            or `INT64` with metadata `"TIMESTAMP_TYPE=NANO/MILLI"`.
        time_assignment:
            Specifies whether the timestamps for the ticks are set to either the start time or end time of the query.
        file_contents:
            If not empty, this method treats its value as a stream from which it should read and construct ticks.
        format_file:
            Specifies the absolute path of the message format file. This parameter applies only to *json* format.
        config_dir:
            Specifies the directory of the normalization file. This parameter applies only to *json* format.
        start:
            Custom start time of the query.
            By default, the start time used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        end:
            Custom end time of the query.
            By default, the start time used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        db: str
            Custom database name for the node of the graph.
            By default, the database used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        tick_type: str
            Custom tick type for the node of the graph.
            By default, "ANY" tick type will be set.
        symbols: str or list of str
            Custom symbol name for the node of the graph.
            By default, the symbol name used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        schema: dict
            Set the schema of the python :py:class:`~onetick.py.Source` object of this class.

            Schema can't be automatically derived from the file, so it should be set manually
            for Python-level type checking to work.
        kwargs:
            Deprecated. Use ``schema`` instead.
            Set the schema of the python :py:class:`~onetick.py.Source` object of this class.

        See also
        --------
        **DATA_FILE_QUERY** OneTick event processor

        Examples
        --------

        Get data from the arrow file:

        .. testcode::
           :skipif: not is_data_file_query_supported()

           import os
           path_to_arrow_file = os.path.join(csv_path, 'data.arrow')

           data = otp.DataFile(path_to_arrow_file)
           df = otp.run(data)
           print(df)

        .. testoutput::

                   Time  A SYMBOL_NAME              T_TIME
           0 2003-12-04  1        AAPL 2003-12-01 00:00:00
           1 2003-12-04  3        AAPL 2003-12-01 02:00:00

        The ``symbol_name_field`` parameter can be used to specify the name of the field containing symbol names.
        The default value for this parameter is **SYMBOL_NAME**, and this field must be specified in the Arrow file.
        The symbols specified for the query will determine which data will be queried from the Arrow file:

        .. testcode::
           :skipif: not is_data_file_query_supported()

           data = otp.DataFile(path_to_arrow_file)
           df = otp.run(data, symbols='MSFT')
           print(df)

        .. testoutput::

                   Time  A SYMBOL_NAME              T_TIME
           0 2003-12-04  2        MSFT 2003-12-01 01:00:00

        To get all symbols from file, you can specify a database without a symbol when running query:

        .. testcode::
           :skipif: not is_data_file_query_supported()

           data = otp.DataFile(path_to_arrow_file)
           df = otp.run(data, symbols=f'{otp.config.default_db}::')
           print(df)

        .. testoutput::

                   Time  A SYMBOL_NAME              T_TIME
           0 2003-12-04  1        AAPL 2003-12-01 00:00:00
           1 2003-12-04  2        MSFT 2003-12-01 01:00:00
           2 2003-12-04  3        AAPL 2003-12-01 02:00:00

        Default time assigned to ticks is query end time.
        Use parameter ``time_assignment`` to change the timestamps for ticks to the query start time:

        .. testcode::
           :skipif: not is_data_file_query_supported()

           data = otp.DataFile(path_to_arrow_file, time_assignment='start')
           df = otp.run(data)
           print(df)

        .. testoutput::

                   Time  A SYMBOL_NAME              T_TIME
           0 2003-12-01  1        AAPL 2003-12-01 00:00:00
           1 2003-12-01  3        AAPL 2003-12-01 02:00:00

        Or use parameter ``timestamp_column`` to get the timestamps from some field:

        .. testcode::
           :skipif: not is_data_file_query_supported()

           data = otp.DataFile(path_to_arrow_file, timestamp_column='T_TIME')
           df = otp.run(data)
           print(df)

        .. testoutput::

                            Time  A SYMBOL_NAME              T_TIME
           0 2003-12-01 00:00:00  1        AAPL 2003-12-01 00:00:00
           1 2003-12-01 02:00:00  3        AAPL 2003-12-01 02:00:00

        The schema of the data can't be checked when constructing query at the Python level,
        so in order to use the fields of the source in other onetick-py methods,
        schema must be specified manually:

        .. testcode::
           :skipif: not is_data_file_query_supported()

           data = otp.DataFile(path_to_arrow_file,
                               schema={'A': int, 'SYMBOL_NAME': str, 'T_TIME': otp.nsectime})
           data['B'] = data['A'] * 2
           df = otp.run(data)
           print(df)

        .. testoutput::

                   Time  A SYMBOL_NAME              T_TIME  B
           0 2003-12-04  1        AAPL 2003-12-01 00:00:00  2
           1 2003-12-04  3        AAPL 2003-12-01 02:00:00  6
        """
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if not is_data_file_query_supported():
            raise RuntimeError("The otp.DataFile is not supported on this OneTick build")

        if not file and not file_contents:
            raise ValueError("One of the parameters 'file' or 'file_contents' must be specified")

        if msg_format not in {'json', 'arrow'}:
            raise ValueError(f"Unknown value for the parameter 'msg_format': {msg_format}")

        if timestamp_column and time_assignment is not utils.adaptive:
            raise ValueError("The parameters 'timestamp_column' and 'time_assignment' cannot be set simultaneously")

        if time_assignment is utils.adaptive:
            time_assignment = 'end'

        if time_assignment not in {'start', 'end'}:
            raise ValueError(f"Unknown value for parameter 'time_assignment': {time_assignment}")

        if msg_format != 'json' and (format_file or config_dir):
            raise ValueError("The parameters 'format_file' and 'config_dir' can only be specified with 'json' format")

        self._file = file
        self._msg_format = msg_format.upper()
        self._symbol_name_field = symbol_name_field
        self._symbology = symbology
        self._timestamp_column = timestamp_column
        self._time_assignment = f'_{time_assignment.upper()}_TIME'
        self._file_contents = file_contents
        self._format_file = format_file
        self._config_dir = config_dir
        self._db = db
        self._tick_type = tick_type

        super().__init__(
            _symbols=symbols, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(schema=schema, **kwargs),
            schema=schema, **kwargs,
        )

    def base_ep(self, schema=None, **kwargs):

        ep_kwargs = {}
        if self._symbology is not None:
            if is_data_file_query_symbology_supported(throw_warning=True,
                                                      feature_name="parameter 'symbology' in otp.DataFile"):
                ep_kwargs['symbology'] = self._symbology

        src = Source(
            otq.DataFileQuery(
                file=self._file or '',
                msg_format=self._msg_format,
                symbol_name_field=self._symbol_name_field,
                timestamp_column=self._timestamp_column or '',
                time_assignment=self._time_assignment,
                file_contents=self._file_contents or '',
                format_file=self._format_file or '',
                config_dir=self._config_dir or '',
                **ep_kwargs,
            ),
            schema=schema,
            **kwargs,
        )

        update_node_tick_type(src, self._tick_type, self._db)

        return src



import weakref
from locator_parser.common import Reader, Writer


class FileWriter(Writer):

    def __init__(self, file_path):
        self.__path = file_path

        super().__init__()

    def flush(self):
        with open(self.__path, "w") as fout:
            for line in self.lines:
                if not line.endswith("\n"):
                    line += "\n"

                fout.write(line)


class PrintWriter(Writer):

    def flush(self):
        print("\n".join(self.lines))


class FileReader(Reader):

    def __init__(self, file_path):
        # pylint: disable-next=consider-using-with
        self.iterable_object = open(file_path.replace('"', ""), "r")
        # pylint: disable-next=unused-private-member
        self.__finalizer = weakref.finalize(self, lambda x: x.close(), self.iterable_object)

        super().__init__()


class LinesReader(Reader):

    def __init__(self, lines):
        self.iterable_object = lines.split("\n")

        super().__init__()



# pylama:ignore=E402,W0611
import os
from . import _version
__version__ = _version.VERSION
__webapi__: bool = bool(os.getenv("OTP_WEBAPI"))


def __validate_onetick_query_integration():  # noqa
    """Logic that checks correctness of integration of python with onetick.query and/or onetick.query_webapi.
    One of the modules should be installed and available to import.

    We first try to import onetick.query and NumPy_OneTickQuery, and if it fails,
    before raising an exception, we check if onetick.query_webapi is installed.
    If it is installed, we set OTP_WEBAPI environment variable and avoid raising any exception/warning.
    """
    global __webapi__

    if os.getenv("OTP_SKIP_OTQ_VALIDATION"):
        return

    _otq_import_ex = None

    try:
        # this import will fail if onetick python directory is not in sys.path
        import onetick.query  # noqa
        try:
            # this import will fail if numpy directory is not in sys.path on old OneTick versions
            import NumPy_OneTickQuery
        except ModuleNotFoundError as e:
            _otq_import_ex = e
        except Exception:
            pass
    except ImportError as e:
        _otq_import_ex = e

    import sysconfig

    config_vars = sysconfig.get_config_vars()

    ot_bin_path = os.path.join("bin")
    ot_python_path = os.path.join(ot_bin_path, "python")
    ot_numpy_path = os.path.join(
        ot_bin_path,
        "numpy",
        "python"
        # short python version 27, 36, 37, etc
        + config_vars["py_version_nodot"]
        # suffix at the end, either empty string for python with the standard memory allocator,
        # or 'm' for python with the py-malloc allocator
        + config_vars["abiflags"],
    )

    pythonpath = os.environ.get('PYTHONPATH')

    missed_required = __get_missed_paths(pythonpath, ot_bin_path, ot_python_path)
    missed_optional = __get_missed_paths(pythonpath, ot_numpy_path)

    import warnings
    if len(missed_required + missed_optional) != 3:
        # TODO: move to onetick.py.compatibility
        if _otq_import_ex is None and onetick.query.OneTickLib.get_build_number() >= 20230711120000:
            warnings.warn(
                'Using PYTHONPATH to specify the location of OneTick python libraries is deprecated.'
                ' Starting from OneTick build 20230711 onetick-py is also distributed as the part of the build,'
                ' and it will override onetick-py you may have installed from pip.'
                ' Use MAIN_ONE_TICK_DIR instead.'
            )

    try:
        from .. import __search_main_one_tick_dir
        main_one_tick_dirs = __search_main_one_tick_dir()
    except ImportError:
        # this exception will be raised if PYTHONPATH was used to specify the location of OneTick libraries
        # and onetick-py doesn't exist in the specified OneTick or have older version than these lines
        main_one_tick_dirs = None

    if _otq_import_ex is None:
        return

    # webapi is installed to pip and OTP_WEBAPI=1, we don't need to raise any exception
    try:
        import onetick.query_webapi
        __webapi__ = True
        return
    except ImportError:
        print('onetick.query_webapi is not available')

    if isinstance(_otq_import_ex, ImportError):
        raise ImportError(
            'The above exception is probably raised '
            'because your python version is unsupported in this OneTick build'
        ) from _otq_import_ex

    # numpy directory is not needed on the latest onetick versions
    # so checking it only if we couldn't import library
    missed_paths = missed_required or missed_optional

    if missed_paths or not main_one_tick_dirs:

        if not missed_paths:
            missed_paths = [ot_bin_path, ot_python_path, ot_numpy_path]

        missed_paths = ', '.join(
            "'" + os.path.join("<path-to-OneTick-dist>", p) + "'"
            for p in missed_paths
        )
        if __webapi__:
            raise ImportError(
                "Environment variable OTP_WEBAPI is set, but we can't import onetick.query_webapi.\n"
                "because we can't find it and related libraries.\n"
                "Please, make sure you installed it with pip.\n"
                "If you want to use OneTick distributed version\n"
                f", make sure that directories {missed_paths} exist and:\n"
                "- either those directories are specified in PYTHONPATH,\n"
                "- or <path-to-OneTick-dist> directory is specified in MAIN_ONE_TICK_DIR (recommended)."
            )

        raise ImportError(
            "We can't import onetick.query, because we can't find it and related libraries.\n"
            f"Please, make sure that directories {missed_paths} exist and:\n"
            "- either those directories are specified in PYTHONPATH,\n"
            "- or <path-to-OneTick-dist> directory is specified in MAIN_ONE_TICK_DIR (recommended)."
        ) from _otq_import_ex

    raise _otq_import_ex


def __get_missed_paths(env, *paths, found_callback=None):
    from pathlib import Path

    missed_paths = []
    env = env or ''
    global_python_path = env.split(os.pathsep)
    for p in paths:
        match = False
        for gb in global_python_path:
            if Path(gb).match(f"*/{p}"):
                match = True
                if found_callback:
                    found_callback(gb, p)
                break
        if not match:
            missed_paths.append(p)
    return missed_paths


def __set_onetick_path_variables():
    if __webapi__ or os.getenv("OTP_SKIP_OTQ_VALIDATION"):
        return
    import warnings
    import onetick.query as otq
    from pathlib import Path

    onetick_query_init = Path(otq.__file__)

    # pylint: disable=global-variable-undefined
    global __main_one_tick_dir__, __one_tick_bin_dir__, __one_tick_python_dir__
    __one_tick_python_dir__ = str(onetick_query_init.parents[2])
    __one_tick_bin_dir__ = str(onetick_query_init.parents[3])
    __main_one_tick_dir__ = str(onetick_query_init.parents[4])

    if os.name == 'nt' and __one_tick_bin_dir__ not in os.environ.get('PATH', '').split(os.pathsep):
        warnings.warn(
            f'OneTick {__one_tick_bin_dir__} is not specified in PATH environment variable.'
            ' Some functionality, e.g. using per-tick script, may be unavailable in this case.'
        )


__validate_onetick_query_integration()
__set_onetick_path_variables()

# -------------------------------------------- #

from ._stack_info import _modify_stack_info_in_onetick_query
_modify_stack_info_in_onetick_query()

# -------------------------------------------- #

from onetick.py import functions, aggregations
from onetick.py.functions import (
    concat, join, join_by_time, apply_query, apply, cut, qcut, merge, coalesce, corp_actions, format,
    join_with_aggregated_window
)
from onetick.py.types import (
    msectime, nsectime, string, nan, inf, datetime, date, dt, varstring,
    _int as int, uint, long, ulong, byte, short, decimal,
    Year, Quarter, Month, Week, Day, Hour, Minute, Second, Milli, Nano,
    default_by_type, timedelta,
)
from onetick.py.sources import (Tick, TTicks, Ticks, Orders, Trades, NBBO, Quotes, Query, CSV, ReadCache, ReadParquet,
                                Custom, query, Symbols, Empty, DataSource, LocalCSVTicks, SymbologyMapping,
                                ObSnapshot, ObSnapshotWide, ObSnapshotFlat, ObSummary, ObSize, ObVwap, ObNumLevels,
                                by_symbol, ODBC, SplitQueryOutputBySymbol, DataFile, PointInTime,
                                ReadSnapshot, ShowSnapshotList, FindSnapshotSymbols)
from onetick.py.utils import adaptive, range, perf
from onetick.py.session import Session, TestSession, Config, Locator, HTTPSession
from onetick.py.servers import RemoteTS, LoadBalancing, FaultTolerance
from onetick.py.db import DB, RefDB
from onetick.py.db._inspection import databases, derived_databases
from onetick.py.cache import create_cache, delete_cache, modify_cache_config
from onetick.py import state
from onetick.py.core.source import Source, MetaFields
from onetick.py.core.multi_output_source import MultiOutputSource
from onetick.py.core.eval_query import eval  # noqa: it is ok to redefine built-in function
from onetick.py.core.per_tick_script import (
    # TODO: wanna access per_tick_script classes as otp.script.static and so on
    Static as static,
    TickDescriptorFields as tick_descriptor_fields,
    tick_list_tick,
    tick_set_tick,
    tick_deque_tick,
    dynamic_tick,
)
from onetick.py.callback import CallbackBase
from onetick.py.sql import SqlQuery
from onetick.py.run import run
from onetick.py.math import rand, now
from onetick.py.misc import bit_and, bit_or, bit_at, bit_xor, bit_not, hash_code, get_symbology_mapping
from onetick.py.core.column import Column
from onetick.py.core.column_operations.base import Operation, Expr as expr, Raw as raw, OnetickParameter as param
from onetick.py.core.per_tick_script import remote, Once, once, logf, throw_exception
from onetick.py.configuration import config
from onetick.py import oqd
from onetick.py.otq import otli
OneTickLib = otli.OneTickLib

try:
    from pandas import date_range
except Exception:
    pass

meta_fields = MetaFields()


def __getattr__(name):
    # actually, these values are not evaluated on module loading anymore,
    # so they can be used without fear
    # but let's raise deprecation warning anyway
    # to encourage users to use otp.config only
    defaults = dict(
        # lambdas are needed in case some config values are not set,
        # so we don't raise exception on getting any attribute
        DEFAULT_START_TIME=lambda: config.default_start_time,
        DEFAULT_END_TIME=lambda: config.default_end_time,
        DEFAULT_TZ=lambda: config.tz,
        DEFAULT_SYMBOL=lambda: config.default_symbol,
        DEFAULT_DB=lambda: config.default_db,
        DEFAULT_DB_SYMBOL=lambda: config.default_db_symbol,
    )
    if name in defaults:
        import warnings
        warnings.warn(
            f'Using otp.{name} is deprecated, use otp.config.* properties instead.',
            FutureWarning
        )
        return defaults[name]()
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


# aliases
funcs = functions  # type: ignore  # noqa
agg = aggregations  # type: ignore  # noqa


# set pandas default pandas options to show all columns
import pandas as _pd

_pd.set_option("display.max_columns", None)
_pd.set_option("display.expand_frame_repr", False)
_pd.set_option("max_colwidth", None)


from onetick.py.otq import otq as _otq
try:
    __build__ = str(_otq.OneTickLib.get_build_number())
except AttributeError:
    __build__ = 'webapi'

# initialize logger module and import public api
from .log import get_logger

del (_version,
     _otq,
     _pd,
     _modify_stack_info_in_onetick_query,
     otli)



from typing import Optional

from onetick.py import types as ott
from onetick.py import configuration, utils
from onetick.py.core.column_operations.accessors._accessor import _Accessor
from onetick.py.core.column_operations.base import _Operation
from onetick.py.backports import Literal
from onetick.py.docs.utils import alias
from onetick.py.compatibility import is_ilike_supported


def _get_onetick_bool_string(value: bool) -> str:
    if value:
        return '"true"'
    return '"false"'


class _StrAccessor(_Accessor):
    """ Accessor for string functions

    >>> data = otp.Ticks(X=['some string'])
    >>> data["Y"] = data["X"].str.<function_name>() # doctest: +SKIP
    """
    class Formatter(_Operation):
        def __init__(self, dtype, formatter, op_params):

            def op_func(*args, **kwargs):
                return formatter(*args, **kwargs), dtype

            super().__init__(op_func=op_func, op_params=op_params, dtype=dtype)

    def to_datetime(self,
                    format='%Y/%m/%d %H:%M:%S.%J',
                    timezone=None,
                    unit: Optional[Literal['ms', 'ns']] = None):
        """
        Converts the formatted time to the number of nanoseconds (datetime) since 1970/01/01 GMT.

        Parameters
        ----------
        format: str, Operation, Column
            The format might contain any characters, but the following combinations of
            characters have special meanings

            %Y - Year (4 digits)

            %y - Year (2 digits)

            %m - Month (2 digits)

            %d - Day of month (2 digits)

            %H - Hours (2 digits, 24-hour format)

            %I - Hours (2 digits, 12-hour format)

            %M - Minutes (2 digits)

            %S - Seconds (2 digits)

            %J - Nanoseconds (9 digits)

            %p - AM/PM (2 characters)

        timezone: str | Operation | Column
            Timezone. The timezone of the query will be used if no ``timezone`` was passed.

        unit: str, optional
            If set, `format` and `timezone` are ignored.
            If equals to `ns`, constructs a nanosecond-granularity timestamp from a millisecond-granularity
            string. It has the following format: < milliseconds since 1970/01/01 GMT >.< fraction of a millisecond >.
            The fraction might have at most six digits. If the fraction is equal to zero,
            .< fraction of a millisecond > is optional.
            If equals to `ms`, constructs a millisecond-granularity timestamp from a millisecond-granularity
            string. It has the following format: < milliseconds since 1970/01/01 GMT >.

        Returns
        -------
        Operation
            :py:class:`nsectime <onetick.py.types.nsectime>` Operation obtained from the string

        Examples
        --------
        >>> # OTdirective: snippet-name: string.to timestamp;
        >>> data = otp.Tick(X='5/17/22-11:10:56.123456789')
        >>> data['Y'] = data['X'].str.to_datetime('%m/%d/%y-%H:%M:%S.%J', 'Europe/London')
        >>> otp.run(data)
                Time                           X                             Y
        0 2003-12-01  5/17/22-11:10:56.123456789 2022-05-17 06:10:56.123456789

        >>> data = otp.Ticks(A=['1693825877111.002001', '1693825877112'])
        >>> data['NSECTIME_A'] = data['A'].str.to_datetime(unit='ns')
        >>> otp.run(data)
                             Time                     A                    NSECTIME_A
        0 2003-12-01 00:00:00.000  1693825877111.002001 2023-09-04 07:11:17.111002001
        1 2003-12-01 00:00:00.001         1693825877112 2023-09-04 07:11:17.112000000

        >>> data = otp.Tick(A='1693825877111')
        >>> data['MSECTIME_A'] = data['A'].str.to_datetime(unit='ms')
        >>> otp.run(data)
                Time              A              MSECTIME_A
        0 2003-12-01  1693825877111 2023-09-04 07:11:17.111
        """
        if unit is None:
            if timezone is utils.default:
                timezone = configuration.config.tz

            def formatter(column, fmt, tz):
                column_str = ott.value2str(column)
                tz_str, format_str = self._preprocess_tz_and_format(tz, fmt)
                format_str = format_str.replace('%f', '%J')
                return f'parse_nsectime({format_str},{column_str},{tz_str})'

            return _StrAccessor.Formatter(
                op_params=[self._base_column, format, timezone],
                dtype=ott.nsectime,
                formatter=formatter,
            )
        else:
            if unit == 'ns':
                return _StrAccessor.Formatter(
                    op_params=[self._base_column],
                    dtype=ott.nsectime,
                    formatter=lambda column: f'MSEC_STR_TO_NSECTIME({ott.value2str(column)})',
                )
            if unit == 'ms':
                return _StrAccessor.Formatter(
                    op_params=[self._base_column],
                    dtype=ott.msectime,
                    formatter=lambda column: f'GET_MSECS(MSEC_STR_TO_NSECTIME({ott.value2str(column)}))',
                )
            raise ValueError(f'`{unit}` is unsupported value for `unit` parameter')

    strptime = alias(to_datetime,
                     doc_replacer=lambda doc: doc.replace('to_datetime', 'strptime'))

    def token(self, sep=" ", n=0):
        """
        Breaks the value into tokens based on the delimiter ``sep``
        and returns token at position ``n`` (zero-based).

        If there are not enough tokens to get the one at position ``n``, then empty string is returned.

        Parameters
        ----------
        sep: str or Column or Operation
            The delimiter, which must be a single character used to split the string into tokens.
        n: int, Operation
            Token index to return. For a negative ``n``, count from the end instead of the beginning.
            If index is out of range, then empty string is returned.

        Returns
        -------
        Operation
            token at position ``n`` or empty string.

        Examples
        -------
        >>> # OTdirective: snippet-name: string.token;
        >>> data = otp.Tick(X='US_COMP::TRD')
        >>> data['Y'] = data['X'].str.token(':', -1)
        >>> otp.run(data)
                Time              X    Y
        0 2003-12-01  US_COMP::TRD  TRD

        Other columns can be used as parameters too:

        >>> data = otp.Tick(X='US_COMP::TRD', SEP=':', N=-1)
        >>> data['Y'] = data['X'].str.token(data['SEP'], data['N'])
        >>> otp.run(data)
                Time              X SEP  N    Y
        0 2003-12-01  US_COMP::TRD   : -1  TRD

        If index is out of range, then empty string is returned:

        >>> data = otp.Tick(X='US_COMP::TRD')
        >>> data['Y'] = data['X'].str.token(':', 999)
        >>> otp.run(data)
                Time              X  Y
        0 2003-12-01  US_COMP::TRD
        """
        if isinstance(sep, str) and len(sep) != 1:
            raise ValueError("Function '.str.token()' expects parameter 'sep' to be a single character")
        return _StrAccessor.Formatter(
            op_params=[self._base_column, sep, n],
            dtype=self._base_column.dtype,
            formatter=lambda column, sep, n: f'token({ott.value2str(column)},{ott.value2str(n)},{ott.value2str(sep)})'
        )

    def match(self, pat, case=True):
        r"""
        Match the text against a regular expression specified in the ``pat`` parameter.

        Parameters
        ----------
        pat: str or Column or Operation
            A pattern specified via the POSIX extended regular expression syntax.
        case: bool
            If ``True``, then regular expression is case-sensitive.

        Returns
        -------
        Operation
            ``True`` if the match was successful, ``False`` otherwise.
            Note that boolean Operation is converted to float if added as a column.

        Examples
        --------
        >>> # OTdirective: snippet-name: string.match;
        >>> data = otp.Ticks(X=['hello', 'there were 77 ticks'])
        >>> data['Y'] = data['X'].str.match(r'\d\d')
        >>> otp.run(data)
                             Time                    X    Y
        0 2003-12-01 00:00:00.000                hello  0.0
        1 2003-12-01 00:00:00.001  there were 77 ticks  1.0

        Other columns can be used as parameter ``pat`` too:

        >>> data = otp.Tick(X='OneTick', PAT='onetick')
        >>> data['Y'] = data['X'].str.match(data['PAT'], case=False)
        >>> otp.run(data)
                Time        X      PAT    Y
        0 2003-12-01  OneTick  onetick  1.0

        ``match`` function can also be used as a filter.
        For example, to filter on-exchange continuous trading trades:

        >>> q = otp.DataSource('US_COMP', tick_type='TRD', symbols=['SPY'])  # doctest: +SKIP
        >>> q = q[['PRICE', 'SIZE', 'COND', 'EXCHANGE']]  # doctest: +SKIP
        >>> q, _ = q[q['COND'].str.match('^[^O6TUHILNRWZ47QMBCGPV]*$')]  # doctest: +SKIP
        >>> otp.run(q, start=otp.dt(2023, 5, 15, 9, 30), end=otp.dt(2023, 5, 15, 9, 30, 1))  # doctest: +SKIP
                                    Time    PRICE  SIZE  COND EXCHANGE
        0  2023-05-15 09:30:00.000776704  412.220   247              Z
        1  2023-05-15 09:30:00.019069440  412.230   100   F          K
        ..                           ...      ...   ...   ...      ...
        """
        caseless = _get_onetick_bool_string(not case)
        return _StrAccessor.Formatter(
            op_params=[self._base_column, pat],
            dtype=bool,
            formatter=lambda column, pat: f'regex_match({ott.value2str(column)},{ott.value2str(pat)},{caseless})',
        )

    def len(self):
        """
        Get the length of a string.

        Returns
        -------
        Operation
            The length of the string.
            If a null-character (byte with value ``0``) is present in the string,
            its position (0-based) is returned.

        Examples
        --------
        >>> # OTdirective: snippet-name: string.len;
        >>> data = otp.Ticks(X=['hello', 'world!'])
        >>> data['LEN'] = data['X'].str.len()
        >>> otp.run(data)
                             Time       X  LEN
        0 2003-12-01 00:00:00.000   hello    5
        1 2003-12-01 00:00:00.001  world!    6
        """
        return _StrAccessor.Formatter(op_params=[self._base_column],
                                      dtype=int,
                                      formatter=lambda column: f'strlen({ott.value2str(column)})')

    def contains(self, substr):
        """
        Check if the string contains ``substr``.

        Note
        ----
        This function does not support regular expressions.
        Use :func:`match` for this purpose.

        Parameters
        ----------
        substr: str or Column or Operation
            A substring to search for within the string.

        Returns
        -------
        Operation
            ``True`` if the string contains the substring, ``False`` otherwise.
            Note that boolean Operation is converted to float if added as a column.

        Examples
        --------
        >>> # OTdirective: snippet-name: string.contains;
        >>> data = otp.Ticks(X=['hello', 'world!'])
        >>> data['CONTAINS'] = data['X'].str.contains('hel')
        >>> otp.run(data)
                             Time       X  CONTAINS
        0 2003-12-01 00:00:00.000   hello       1.0
        1 2003-12-01 00:00:00.001  world!       0.0

        Other columns can be used as parameter ``substr`` too:

        >>>  # OTdirective: snippet-name: string.contains another field;
        >>> data = otp.Ticks(X=['hello', 'big', 'world!'],
        ...                  Y=['hel', 'wor', 'wor'])
        >>> data['CONTAINS'] = data['X'].str.contains(data['Y'])
        >>> otp.run(data)
                             Time       X    Y  CONTAINS
        0 2003-12-01 00:00:00.000   hello  hel       1.0
        1 2003-12-01 00:00:00.001     big  wor       0.0
        2 2003-12-01 00:00:00.002  world!  wor       1.0

        This method can also be used for filtering:

        >>>  # OTdirective: snippet-name: string.contains as a filter;
        >>> data = otp.Ticks(X=['Hello', 'World'])
        >>> with_substr, wo_substr = data[data['X'].str.contains('Hel')]
        >>> otp.run(with_substr)
                Time      X
        0 2003-12-01  Hello
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, substr],
            dtype=bool,
            formatter=lambda column, substr: f'instr({ott.value2str(column)}, {ott.value2str(substr)}) > -1',
        )

    def trim(self):
        """
        Removes white spaces from both sides of the string.

        Returns
        -------
        Operation
            Trimmed string

        See Also
        --------
        :meth:`ltrim`, :meth:`rtrim`

        Examples
        --------
        >>> # OTdirective: snippet-name: string.trim;
        >>> data = otp.Ticks(X=['  Hello', 'World  '])
        >>> data['X'] = data['X'].str.trim()
        >>> otp.run(data)
                             Time      X
        0 2003-12-01 00:00:00.000  Hello
        1 2003-12-01 00:00:00.001  World
        """
        return _StrAccessor.Formatter(op_params=[self._base_column],
                                      dtype=self._base_column.dtype,
                                      formatter=lambda column: f'trim({ott.value2str(column)})')

    def ltrim(self):
        """
        Removes the leading white spaces from a string.

        Returns
        -------
        Operation
            Trimmed string

        See Also
        --------
        :meth:`trim`, :meth:`rtrim`
        """
        return _StrAccessor.Formatter(op_params=[self._base_column],
                                      dtype=self._base_column.dtype,
                                      formatter=lambda column: f'ltrim({ott.value2str(column)})')

    def rtrim(self):
        """
        Removes the trailing white spaces from a string.

        Returns
        -------
        Operation
            Trimmed string

        See Also
        --------
        :meth:`ltrim`, :meth:`trim`
        """
        return _StrAccessor.Formatter(op_params=[self._base_column],
                                      dtype=self._base_column.dtype,
                                      formatter=lambda column: f'rtrim({ott.value2str(column)})')

    def lower(self):
        """
        Convert a string to lower case.

        Returns
        -------
        Operation
            Lowercased string

        Examples
        --------
        >>> # OTdirective: snippet-name: string.lower;
        >>> data = otp.Ticks(X=['HeLlO', 'wOrLd!'])
        >>> data['LOW'] = data['X'].str.lower()
        >>> otp.run(data)
                             Time       X     LOW
        0 2003-12-01 00:00:00.000   HeLlO   hello
        1 2003-12-01 00:00:00.001  wOrLd!  world!
        """
        return _StrAccessor.Formatter(op_params=[self._base_column],
                                      dtype=self._base_column.dtype,
                                      formatter=lambda column: f'lower({ott.value2str(column)})')

    def upper(self):
        """
        Converts a string to upper case.

        Returns
        -------
        Operation
            Uppercased string

        Examples
        --------
        >>> # OTdirective: snippet-name: string.upper;
        >>> data = otp.Ticks(X=['HeLlO', 'wOrLd!'])
        >>> data['UP'] = data['X'].str.upper()
        >>> otp.run(data)
                             Time       X      UP
        0 2003-12-01 00:00:00.000   HeLlO   HELLO
        1 2003-12-01 00:00:00.001  wOrLd!  WORLD!
        """
        return _StrAccessor.Formatter(op_params=[self._base_column],
                                      dtype=self._base_column.dtype,
                                      formatter=lambda column: f'upper({ott.value2str(column)})')

    def replace(self, pat, repl):
        """
        Search for occurrences (case dependent) of ``pat`` and replace with ``repl``.

        Parameters
        ----------
        pat: str or Column or Operation
            Pattern to replace.
        repl: str or Column or Operation
            Replacement string.

        Returns
        -------
        Operation
            String with ``pat`` replaced by ``repl``.

        Examples
        --------
        >>> # OTdirective: snippet-name: string.replace;
        >>> data = otp.Ticks(X=['A Table', 'A Chair', 'An Apple'])
        >>> data['Y'] = data['X'].str.replace('A', 'The')
        >>> otp.run(data)
                             Time         X             Y
        0 2003-12-01 00:00:00.000   A Table     The Table
        1 2003-12-01 00:00:00.001   A Chair     The Chair
        2 2003-12-01 00:00:00.002  An Apple  Then Thepple

        Other columns can be used as parameters too:

        >>> # OTdirective: snippet-name: string.replace from field;
        >>> data = otp.Ticks(X=['A Table', 'A Chair', 'An Apple'],
        ...                  PAT=['A', 'A', 'An'],
        ...                  REPL=['The', 'Their', 'My'])
        >>> data['Y'] = data['X'].str.replace(data['PAT'], data['REPL'])
        >>> otp.run(data)
                             Time         X PAT   REPL            Y
        0 2003-12-01 00:00:00.000   A Table   A    The    The Table
        1 2003-12-01 00:00:00.001   A Chair   A  Their  Their Chair
        2 2003-12-01 00:00:00.002  An Apple  An     My     My Apple
        """
        # see, BDS-112
        if not isinstance(pat, str):
            pat = pat.str.rtrim()
        if not isinstance(repl, str):
            repl = repl.str.rtrim()
        return _StrAccessor.Formatter(
            op_params=[self._base_column, pat, repl],
            dtype=self._base_column.dtype,
            formatter=(
                lambda column, pat, repl:
                f'replace({ott.value2str(column)}, {ott.value2str(pat)}, {ott.value2str(repl)})'
            ),
        )

    def regex_replace(self, pat, repl, *, replace_every=False, caseless=False):
        r"""
        Search for occurrences (case dependent) of ``pat`` and replace with ``repl``.

        Parameters
        ----------
        pat: str or Column or Operation
            Pattern to replace specified via the POSIX extended regular expression syntax.
        repl: str or Column or Operation
            Replacement string. ``\0`` refers to the entire matched text. ``\1`` to ``\9`` refer
            to the text matched by the corresponding parenthesized group in ``pat``.
        replace_every: bool
            If ``replace_every`` flag is set to ``True``, all matches will be replaced, if ``False`` only the first one.
        caseless: bool
            If the ``caseless`` flag is set to ``True``, matching is case-insensitive.

        Returns
        -------
        Operation
            String with pattern ``pat`` replaced by ``repl``.

        See Also
        --------
        :meth:`extract`

        Examples
        --------
        >>> # OTdirective: snippet-name: string.regex replace;
        >>> data = otp.Ticks(X=['A Table', 'A Chair', 'An Apple'])
        >>> data['Y'] = data['X'].str.regex_replace('An? ', 'The ')
        >>> otp.run(data)
                             Time         X          Y
        0 2003-12-01 00:00:00.000   A Table  The Table
        1 2003-12-01 00:00:00.001   A Chair  The Chair
        2 2003-12-01 00:00:00.002  An Apple  The Apple

        Parameter ``replace_every`` will replace all occurrences of ``pat`` in the string:

        >>> # OTdirective: snippet-name: string.regex replace all;
        >>> data = otp.Ticks(X=['A Table, A Chair, An Apple'])
        >>> data['Y'] = data['X'].str.regex_replace('An? ', 'The ', replace_every=True)
        >>> otp.run(data)
                Time                           X                                Y
        0 2003-12-01  A Table, A Chair, An Apple  The Table, The Chair, The Apple

        Capturing groups in regular expressions is supported:

        >>> # OTdirective: snippet-name: string.regex groups;
        >>> data = otp.Ticks(X=['11/12/1992', '9/22/1993', '3/30/1991'])
        >>> data['Y'] = data['X'].str.regex_replace(r'(\d{1,2})/(\d{1,2})/', r'\2.\1.')
        >>> otp.run(data)
                             Time           X           Y
        0 2003-12-01 00:00:00.000  11/12/1992  12.11.1992
        1 2003-12-01 00:00:00.001   9/22/1993   22.9.1993
        2 2003-12-01 00:00:00.002   3/30/1991   30.3.1991
        """
        replace_every = _get_onetick_bool_string(replace_every)
        caseless = _get_onetick_bool_string(caseless)
        return _StrAccessor.Formatter(
            op_params=[self._base_column, pat, repl],
            dtype=self._base_column.dtype,
            formatter=lambda column, pat, repl: f'regex_replace({ott.value2str(column)}, {ott.value2str(pat)},'
                                                f' {ott.value2str(repl)}, {replace_every}, {caseless})',
        )

    def find(self, sub, start=0):
        """
        Find the index of ``sub`` in the string. If not found, returns ``-1``.

        Parameters
        ----------
        sub: str or Column or Operation
            Substring to find.
        start: int or Column or Operation
            Starting position to find.

        Returns
        -------
        Operation
            The starting position of the substring or ``-1`` if it is not found.

        Examples
        --------
        >>> data = otp.Ticks(X=['ananas', 'banana', 'potato'])
        >>> data['Y'] = data['X'].str.find('ana') # OTdirective: snippet-name: string.find;
        >>> otp.run(data)
                             Time       X  Y
        0 2003-12-01 00:00:00.000  ananas  0
        1 2003-12-01 00:00:00.001  banana  1
        2 2003-12-01 00:00:00.002  potato -1

        Other columns can be used as parameter ``sub`` too:

        >>> # OTdirective: snippet-name: string.find field value;
        >>> data = otp.Ticks(X=['Ananas', 'Banana', 'Potato'], sub=['Ana', 'anan', 'ato'])
        >>> data['Y'] = data['X'].str.find(data['sub'])
        >>> otp.run(data)
                             Time       X   sub  Y
        0 2003-12-01 00:00:00.000  Ananas   Ana  0
        1 2003-12-01 00:00:00.001  Banana  anan  1
        2 2003-12-01 00:00:00.002  Potato   ato  3

        Note that empty string will be found at the start of any string:

        >>> data = otp.Ticks(X=['string', ''])
        >>> data['Y'] = data['X'].str.find('')
        >>> otp.run(data)
                             Time       X  Y
        0 2003-12-01 00:00:00.000  string  0
        1 2003-12-01 00:00:00.001          0

        ``start`` parameter is used to find ``sub`` starting from selected position:

        >>> data = otp.Ticks(X=['ababab', 'abbbbb'])
        >>> data['Y'] = data['X'].str.find('ab', 1)
        >>> otp.run(data)
                             Time       X  Y
        0 2003-12-01 00:00:00.000  ababab  2
        1 2003-12-01 00:00:00.001  abbbbb -1
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, sub, start],
            dtype=int,
            formatter=(
                lambda column, sub, start:
                f'LOCATE({ott.value2str(sub)}, {ott.value2str(column)}, {ott.value2str(start + 1)})-1'
            ),
        )

    def repeat(self, repeats):
        """
        Duplicate a string ``repeats`` times.

        Note
        ----
        * Alternative for the ``repeat`` function is multiplication.
        * The returned string has the same type and maximum length as the original field.

        Parameters
        ----------
        repeats: int or Column or Operation
            Non-negative number of copies of the string.
            Repeating zero times results in empty string.
            Repeating negative number of times results in exception.

        Returns
        -------
        Operation
            String repeated ``repeats`` times.

        Examples
        --------
        >>> # OTdirective: snippet-name: string.repeat;
        >>> data = otp.Ticks(X=['Banana', 'Ananas', 'Apple'])
        >>> data['X'] = data['X'].str.repeat(3)
        >>> otp.run(data)
                             Time                   X
        0 2003-12-01 00:00:00.000  BananaBananaBanana
        1 2003-12-01 00:00:00.001  AnanasAnanasAnanas
        2 2003-12-01 00:00:00.002     AppleAppleApple

        Other columns can be used as parameter ``repeats`` too:

        # OTdirective: snippet-name: string.repeat from a field;
        >>> data = otp.Ticks(X=['Banana', 'Ananas', 'Apple'], TIMES=[1, 3, 2])
        >>> data['Y'] = data['X'].str.repeat(data['TIMES'])
        >>> otp.run(data)
                             Time       X  TIMES                   Y
        0 2003-12-01 00:00:00.000  Banana      1              Banana
        1 2003-12-01 00:00:00.001  Ananas      3  AnanasAnanasAnanas
        2 2003-12-01 00:00:00.002   Apple      2          AppleApple

        The returned string has the same type and therefore the same maximum length as the original field:

        >>> data = otp.Ticks(X=[otp.string[9]('Banana')])
        >>> data['Y'] = data['X'].str.repeat(3)
        >>> data.schema
        {'X': string[9], 'Y': string[9]}
        >>> otp.run(data)
                Time       X          Y
        0 2003-12-01  Banana  BananaBan

        ``repeat`` does the same thing as multiplication by a non-negative int:

        >>> # OTdirective: snippet-name: string.repeat by multiplication;
        >>> data = otp.Ticks(X=['Banana'], N=[2])
        >>> data['X2'] = data['X'] * data['N']
        >>> data['X3'] = data['X'] * 3
        >>> otp.run(data)
                Time       X  N            X2                  X3
        0 2003-12-01  Banana  2  BananaBanana  BananaBananaBanana

        Multiplying by 0 results in empty string:

        >>> data = otp.Ticks(X=['Banana', 'Apple'])
        >>> data['Y'] = data['X'].str.repeat(0)
        >>> otp.run(data)
                             Time       X Y
        0 2003-12-01 00:00:00.000  Banana
        1 2003-12-01 00:00:00.001   Apple
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, repeats],
            dtype=self._base_column.dtype,
            formatter=lambda column, repeats: f'repeat({ott.value2str(column)}, {ott.value2str(repeats)})',
        )

    def extract(self, pat, rewrite=r"\0", caseless=False):
        r"""
        Match the string against a regular expression specified by ``pat`` and return the first match.
        The ``rewrite`` parameter can optionally be used to arrange the matched substrings and embed them within the
        string specified in ``rewrite``.

        Parameters
        ----------
        pat: str or Column or Operation
            Pattern to search for specified via the POSIX extended regular expression syntax.
        rewrite: str or Column or Operation
            A string that specifies how to arrange the matched text. ``\0`` refers to the entire matched text.
            ``\1`` to ``\9`` refer to the text matched by the corresponding parenthesized group in ``pat``.
            ``\u`` and ``\l`` modifiers within the ``rewrite`` string convert the case of the text that
            matches the corresponding parenthesized group (e.g., ``\u1`` converts ``\1`` to uppercase).
        caseless: bool
            If the ``caseless`` flag is set to ``True``, matching is case-insensitive.

        Returns
        -------
        Operation
            String matched by ``pat`` with format specified in ``rewrite``.

        See Also
        --------
        regex_replace

        Examples
        --------
        >>> # OTdirective: snippet-name: string.regex extract;
        >>> data = otp.Ticks(X=['Mr. Smith: +1348 +4781', 'Ms. Smith: +8971'])
        >>> data['TEL'] = data['X'].str.extract(r'\+\d{4}')
        >>> otp.run(data)
                             Time                       X    TEL
        0 2003-12-01 00:00:00.000  Mr. Smith: +1348 +4781  +1348
        1 2003-12-01 00:00:00.001        Ms. Smith: +8971  +8971

        You can specify the group to extract in the ``rewrite`` parameter:

        >>> # OTdirective: snippet-name: string.regex extract group;
        >>> data = otp.Ticks(X=['Mr. Smith: 1992/12/22', 'Ms. Smith: 1989/10/15'])
        >>> data['BIRTH_YEAR'] = data['X'].str.extract(r'(\d{4})/(\d{2})/(\d{2})', rewrite=r'birth year: \1')
        >>> otp.run(data)
                             Time                      X        BIRTH_YEAR
        0 2003-12-01 00:00:00.000  Mr. Smith: 1992/12/22  birth year: 1992
        1 2003-12-01 00:00:00.001  Ms. Smith: 1989/10/15  birth year: 1989

        You can use a column as a ``rewrite`` or ``pat`` parameter:

        >>> # OTdirective: snippet-name: string.regex extract from field;
        >>> data = otp.Ticks(X=['Kelly, Mr. James', 'Wilkes, Mrs. James', 'Connolly, Miss. Kate'],
        ...                  PAT=['(Mrs?)\\.', '(Mrs?)\\.', '(Miss)\\.'],
        ...                  REWRITE=['Title 1: \\1', 'Title 2: \\1', 'Title 3: \\1'])
        >>> data['TITLE'] = data['X'].str.extract(data['PAT'], rewrite=data['REWRITE'])
        >>> otp.run(data)
                             Time                     X       PAT      REWRITE          TITLE
        0 2003-12-01 00:00:00.000      Kelly, Mr. James  (Mrs?)\.  Title 1: \1  Title 1:   Mr
        1 2003-12-01 00:00:00.001    Wilkes, Mrs. James  (Mrs?)\.  Title 2: \1  Title 2:  Mrs
        2 2003-12-01 00:00:00.002  Connolly, Miss. Kate  (Miss)\.  Title 3: \1  Title 3: Miss

        Case of the extracted string can be changed by adding ``l`` and ``u`` to extract group:

        >>> # OTdirective: snippet-name: string.regex extract caseless;
        >>> data = otp.Ticks(NAME=['mr. BroWn', 'Ms. smITh'])
        >>> data['RESULT'] = data['NAME'].str.extract(r'(m)([rs]\. )([a-z])([a-z]*)', r'\u1\l2\u3\l4', caseless=True)
        >>> otp.run(data)
                             Time       NAME     RESULT
        0 2003-12-01 00:00:00.000  mr. BroWn  Mr. Brown
        1 2003-12-01 00:00:00.001  Ms. smITh  Ms. Smith
        """
        caseless = _get_onetick_bool_string(caseless)
        return _StrAccessor.Formatter(
            op_params=[self._base_column, pat, rewrite],
            dtype=self._base_column.dtype,
            formatter=(
                lambda column, pat, rewrite:
                f'regex_extract({ott.value2str(column)}, {ott.value2str(pat)}, {ott.value2str(rewrite)}, {caseless})'
            ),
        )

    def substr(self, start, n_bytes=None, rtrim=False):
        """
        Return ``n_bytes`` characters starting from ``start``.

        For a positive ``start`` return ``num_bytes`` of the string, starting from the position specified by
        ``start`` (0-based).
        For a negative ``start``, the position is counted from the end of the string.
        If the ``n_bytes`` parameter is omitted, returns the part of the input string
        starting at ``start`` till the end of the string.

        Parameters
        ----------
        start: int or Column or Operation
            Index of first symbol in substring
        n_bytes: int or Column or Operation
            Number of bytes in substring
        rtrim: bool
            If set to ``True``, original string will be trimmed from the right side
            before getting the substring, this can be useful with negative ``start`` index.

        Returns
        -------
        Operation
            Substring of string (``n_bytes`` length starting with ``start``).

        Examples
        --------
        >>> # OTdirective: snippet-name: string.substring;
        >>> data = otp.Ticks(X=['abcdef', '12345     '], START_INDEX=[2, 1], N=[2, 3])
        >>> data['FIRST_3'] = data['X'].str.substr(0, 3)
        >>> data['LAST_3'] = data['X'].str.substr(-3, rtrim=True)
        >>> data['CENTER'] = data['X'].str.substr(data['START_INDEX'], data['N'])
        >>> otp.run(data)
                             Time       X  START_INDEX  N FIRST_3 LAST_3 CENTER
        0 2003-12-01 00:00:00.000  abcdef            2  2     abc    def     cd
        1 2003-12-01 00:00:00.001   12345            1  3     123    345    234
        """
        column = self._base_column
        if rtrim:
            column = column.str.rtrim()

        if n_bytes is None:
            return _StrAccessor.Formatter(
                op_params=[column, start],
                dtype=self._base_column.dtype,
                formatter=(
                    lambda column, start:
                    f'substr({ott.value2str(column)}, {ott.value2str(start)})'
                ),
            )
        else:
            return _StrAccessor.Formatter(
                op_params=[column, start, n_bytes],
                dtype=self._base_column.dtype,
                formatter=(
                    lambda column, start, n_bytes:
                    f'substr({ott.value2str(column)}, {ott.value2str(start)}, {ott.value2str(n_bytes)})'
                ),
            )

    def get(self, i):
        """
        Returns the character at the position indicated by the 0-based index; and empty string,
        if position is greater or equal to the length.

        Parameters
        ----------
        i: int or Column or Operation
            Index of the character to find.

        Examples
        --------
        >>> data = otp.Ticks(X=['abcdef', '12345     ', 'qw'], GET_INDEX=[2, 1, 0])
        >>> data['THIRD'] = data['X'].str.get(2)
        >>> data['FROM_INDEX'] = data['X'].str.get(data['GET_INDEX'])
        >>> otp.run(data)
                             Time           X  GET_INDEX THIRD FROM_INDEX
        0 2003-12-01 00:00:00.000      abcdef          2     c          c
        1 2003-12-01 00:00:00.001  12345               1     3          2
        2 2003-12-01 00:00:00.002          qw          0                q

        It is possible to use syntax with indexer to call this method:

        >>> data = otp.Ticks(X=['abcdef', '12345     ', 'qw'])
        >>> data['THIRD'] = data['X'].str[1]
        >>> otp.run(data)
                             Time           X THIRD
        0 2003-12-01 00:00:00.000      abcdef     b
        1 2003-12-01 00:00:00.001  12345          2
        2 2003-12-01 00:00:00.002          qw     w
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, i],
            dtype=str,
            formatter=(
                lambda column, i:
                'CASE(BYTE_AT({0}, {1}),-1,"",CHAR(BYTE_AT({0}, {1})))'.format(ott.value2str(column), ott.value2str(i))
            ),
        )

    def concat(self, other):
        """
        Returns a string that is the result of concatenating to ``others``.

        Parameters
        ----------
        other: str or Column or Operation
            String to concatenate with.

        Examples
        --------
        >>> data = otp.Ticks(X=['X1', 'X2', 'X3'], Y=['Y1', 'Y2', 'Y3'])
        >>> data['X_WITH_CONST_SUFFIX'] = data['X'].str.concat('_suffix')
        >>> data['X_WTH_Y'] = data['X'].str.concat(data['Y'])
        >>> otp.run(data)
                             Time   X   Y X_WITH_CONST_SUFFIX X_WTH_Y
        0 2003-12-01 00:00:00.000  X1  Y1           X1_suffix    X1Y1
        1 2003-12-01 00:00:00.001  X2  Y2           X2_suffix    X2Y2
        2 2003-12-01 00:00:00.002  X3  Y3           X3_suffix    X3Y3
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, other],
            dtype=self._base_column.dtype,
            formatter=lambda column, other: f'CONCAT({ott.value2str(column)}, {ott.value2str(other)})',
        )

    def insert(self, start, length, value):
        """
        Returns a string where ``length`` characters have been deleted from string,
        beginning at ``start``, and where ``value`` has been inserted into string, beginning at ``start``.

        Parameters
        ----------
        start: int or Column or Operation
            Position to remove from and to insert into.
        length: int or Column or Operation
            Number if characters to remove.
        value: str or Column or Operation
            String to insert.

        Examples
        --------
        >>> data = otp.Ticks(X=['aaaaaaa', 'bbbbb', 'cccc'], Y=['ddd', 'ee', 'f'])
        >>> data['INSERTED_1'] = data['X'].str.insert(3, 1, 'X')
        >>> data['INSERTED_2'] = data['X'].str.insert(3, 2, 'X')
        >>> data['INSERTED_Y'] = data['X'].str.insert(3, 2, data['Y'])
        >>> otp.run(data)
                             Time        X    Y INSERTED_1 INSERTED_2 INSERTED_Y
        0 2003-12-01 00:00:00.000  aaaaaaa  ddd    aaXaaaa     aaXaaa   aadddaaa
        1 2003-12-01 00:00:00.001    bbbbb   ee      bbXbb       bbXb      bbeeb
        2 2003-12-01 00:00:00.002     cccc    f       ccXc        ccX        ccf

        It is possible to insert without removal:

        >>> data = otp.Ticks(X=['aaaaaaa', 'bbbbb', 'cccc'])
        >>> data['INSERTED'] = data['X'].str.insert(3, 0, 'X')
        >>> otp.run(data)
                             Time        X  INSERTED
        0 2003-12-01 00:00:00.000  aaaaaaa  aaXaaaaa
        1 2003-12-01 00:00:00.001    bbbbb    bbXbbb
        2 2003-12-01 00:00:00.002     cccc     ccXcc

        It is possible to remove without insertion:

        >>> data = otp.Ticks(X=['aaaaaaa', 'bbbbb', 'cccc'])
        >>> data['REMOVED'] = data['X'].str.insert(3, 2, '')
        >>> otp.run(data)
                             Time        X REMOVED
        0 2003-12-01 00:00:00.000  aaaaaaa   aaaaa
        1 2003-12-01 00:00:00.001    bbbbb     bbb
        2 2003-12-01 00:00:00.002     cccc      cc
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, start, length, value],
            dtype=self._base_column.dtype,
            formatter=(
                lambda column, start, length, value:
                f'INSERT({ott.value2str(column)}, {ott.value2str(start)},'
                f' {ott.value2str(length)}, {ott.value2str(value)})'
            ),
        )

    def first(self, count=1):
        """
        Returns first ``count`` symbols.

        Parameters
        ----------
        count: int or Column or Operation
            Number of first symbols to return. Default: 1

        Examples
        --------
        >>> data = otp.Ticks(X=['abc', 'bac', 'cba'], Y=[3, 1, 10])
        >>> data['FIRST'] = data['X'].str.first()
        >>> data['FIRST_Y'] = data['X'].str.first(data['Y'])
        >>> otp.run(data)
                             Time    X   Y FIRST FIRST_Y
        0 2003-12-01 00:00:00.000  abc   3     a     abc
        1 2003-12-01 00:00:00.001  bac   1     b       b
        2 2003-12-01 00:00:00.002  cba  10     c     cba
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, count],
            dtype=str,
            formatter=lambda column, count: f'LEFT({ott.value2str(column)}, {ott.value2str(count)})',
        )

    def last(self, count=1):
        """
        Returns last ``count`` symbols.

        Parameters
        ----------
        count: int or Column or Operation
            Number of last symbols to return. Default: 1

        Examples
        --------
        >>> data = otp.Ticks(X=['abc', 'bac', 'cba'], Y=[3, 1, 9])
        >>> data['LAST'] = data['X'].str.last()
        >>> data['LAST_Y'] = data['X'].str.last(data['Y'])
        >>> otp.run(data)
                             Time    X  Y LAST LAST_Y
        0 2003-12-01 00:00:00.000  abc  3    c    abc
        1 2003-12-01 00:00:00.001  bac  1    c      c
        2 2003-12-01 00:00:00.002  cba  9    a    cba
        """
        # RIGHT function works strange with negative index
        # RIGHT_UTF8 works fine but it is not supported by old builds
        return _StrAccessor.Formatter(
            op_params=[self._base_column, count],
            dtype=self._base_column.dtype,
            formatter=(
                lambda column, count:
                'SUBSTR({0}, MAX(STRLEN({0})-{1}, 0))'.format(ott.value2str(column), ott.value2str(count))
            ),
        )

    def startswith(self, value):
        """
        Checks if the Operation starts with a string.

        Parameters
        ----------
        value: str or Column or Operation
            String to check if starts with it.

        Examples
        --------
        >>> data = otp.Ticks(X=['baaaa', 'bbbbb', 'cbbc'], Y=['ba', 'abb', 'c'])
        >>> data['STARTSWITH_CONST'] = data['X'].str.startswith('bb')
        >>> data['STARTSWITH_Y'] = data['X'].str.startswith(data['Y'])
        >>> otp.run(data)
                             Time      X    Y  STARTSWITH_CONST  STARTSWITH_Y
        0 2003-12-01 00:00:00.000  baaaa   ba               0.0           1.0
        1 2003-12-01 00:00:00.001  bbbbb  abb               1.0           0.0
        2 2003-12-01 00:00:00.002   cbbc    c               0.0           1.0
        """
        return _StrAccessor.Formatter(
            op_params=[self._base_column, value],
            dtype=bool,
            formatter=(
                lambda column, value:
                'LEFT({0}, STRLEN({1}))={1}'.format(ott.value2str(column), ott.value2str(value))
            ),
        )

    def endswith(self, value):
        """
        Checks if the Operation ends with a string.

        Parameters
        ----------
        value: str or Column or Operation
            String to check if starts with it.

        Examples
        --------
        >>> data = otp.Ticks(X=['baaaa', 'bbbbb', 'cbbc', 'c'], Y=['ba', 'bbb', 'c', 'cc'])
        >>> data['ENDSWITH_CONST'] = data['X'].str.endswith('bb')
        >>> data['ENDSWITH_Y'] = data['X'].str.endswith(data['Y'])
        >>> otp.run(data)
                             Time      X    Y  ENDSWITH_CONST  ENDSWITH_Y
        0 2003-12-01 00:00:00.000  baaaa   ba             0.0         0.0
        1 2003-12-01 00:00:00.001  bbbbb  bbb             1.0         1.0
        2 2003-12-01 00:00:00.002   cbbc    c             0.0         1.0
        3 2003-12-01 00:00:00.003      c   cc             0.0         0.0
        """
        # RIGHT function works strange with negative index
        # RIGHT_UTF8 works fine but it is not supported by old builds
        return _StrAccessor.Formatter(
            op_params=[self._base_column, value],
            dtype=bool,
            formatter=(
                lambda column, value:
                'SUBSTR({0}, MAX(STRLEN({0})-STRLEN({1}), 0))={1}'.format(ott.value2str(column), ott.value2str(value))
            ),
        )

    def slice(self, start=None, stop=None):
        """
        Returns slice.

        Parameters
        ----------
        start: int or Column or Operation, optional
            Start position for slice operation.
        stop: int or Column or Operation, optional
            Stop position for slice operation.

        Examples
        --------
        >>> data = otp.Ticks(X=['12345', 'abcde', 'qwerty'], START=[3, 0, 1], STOP=[4, 3, 3])
        >>> data['START_1_SLICE'] = data['X'].str.slice(start=1)
        >>> data['STOP_2_SLICE'] = data['X'].str.slice(stop=2)
        >>> data['SLICE_FROM_COLUMNS'] = data['X'].str.slice(start=data['START'], stop=data['STOP'])
        >>> otp.run(data)
                                     Time       X  START  STOP START_1_SLICE STOP_2_SLICE SLICE_FROM_COLUMNS
        0 2003-12-01 00:00:00.000   12345      3     4          2345           12                  4
        1 2003-12-01 00:00:00.001   abcde      0     3          bcde           ab                abc
        2 2003-12-01 00:00:00.002  qwerty      1     3         werty           qw                 we

        Parameters can be negative:

        >>> data = otp.Ticks(X=['12345', 'abcde', 'qwerty'])
        >>> data['START_SLICE'] = data['X'].str.slice(start=-3)
        >>> data['STOP_SLICE'] = data['X'].str.slice(stop=-1)
        >>> data['START_STOP_SLICE'] = data['X'].str.slice(start=-3, stop=-1)
        >>> otp.run(data)
                             Time       X START_SLICE STOP_SLICE START_STOP_SLICE
        0 2003-12-01 00:00:00.000   12345         345       1234               34
        1 2003-12-01 00:00:00.001   abcde         cde       abcd               cd
        2 2003-12-01 00:00:00.002  qwerty         rty      qwert               rt

        It is possible to use syntax with indexer to call this method:

        >>> data = otp.Ticks(X=['12345', 'abcde', 'qwerty'])
        >>> data['START_SLICE'] = data['X'].str[1:]
        >>> data['STOP_SLICE'] = data['X'].str[:3]
        >>> data['START_STOP_SLICE'] = data['X'].str[1:3]
        >>> otp.run(data)
                             Time       X START_SLICE STOP_SLICE START_STOP_SLICE
        0 2003-12-01 00:00:00.000   12345        2345        123               23
        1 2003-12-01 00:00:00.001   abcde        bcde        abc               bc
        2 2003-12-01 00:00:00.002  qwerty       werty        qwe               we
        """
        if start is None and stop is None:
            raise ValueError("At least one of the `start` or `stop` parameters should be set.")
        if start is None:
            def formatter(x, start, stop):
                x = ott.value2str(x)
                stop_str = ott.value2str(stop)
                len_x = f'STRLEN({x})'
                return (f'CASE({stop_str}>=0,1,'
                        f'SUBSTR({x},0,{stop_str}),'
                        f'SUBSTR({x},0,MAX(0,{len_x}+{stop_str})))')
        elif stop is None:
            def formatter(x, start, stop):
                x = ott.value2str(x)
                len_x = f'STRLEN({x})'
                # we need this workaround because simple RIGHT and SUBSTR with negative start parameter work strange
                # SUBSTR_UTF8 works fine but it is not supported by old builds
                x_corrected = f'LEFT({x},{len_x})'
                # SUBSTR returns '' when ABC(second parameter) >= STRLEN
                return f'SUBSTR({x_corrected},MAX({ott.value2str(start)},-{len_x}))'
        else:
            def formatter(x, start, stop):
                x = ott.value2str(x)
                stop_str = ott.value2str(stop)
                len_x = f'STRLEN({x})'
                # we need this workaround because simple RIGHT and SUBSTR with negative start parameter work strange
                # SUBSTR_UTF8 works fine but it is not supported by old builds
                x_corrected = f'LEFT({x},{len_x})'
                # y is x after cutting the left part (we need to cut the right part of it)
                # SUBSTR returns '' when ABC(second parameter) >= STRLEN
                y = f'SUBSTR({x_corrected},MAX({ott.value2str(start)},-{len_x}))'
                len_y = f'STRLEN({y})'
                len_cut = f'({len_x}-{len_y})'  # length of already cut part (the left one)
                stop_for_y = f'CASE({stop_str}>=0,1,{stop_str}-{len_cut},{stop_str})'
                return (f'CASE({stop_for_y}>=0,1,'
                        f'SUBSTR({y},0,{stop_for_y}),'
                        f'SUBSTR({y},0,MAX(0,{len_y}+{stop_for_y})))')
        return _StrAccessor.Formatter(op_params=[self._base_column, start, stop],
                                      dtype=self._base_column.dtype,
                                      formatter=formatter)

    def __getitem__(self, item):
        if isinstance(item, slice):
            if item.step is not None:
                raise ValueError("`step` parameter is not supported.")
            return self.slice(start=item.start, stop=item.stop)
        return self.get(item)

    def like(self, pattern):
        r"""
        Check if the value is matched with SQL-like ``pattern``.

        Parameters
        ----------
        pattern: str or symbol parameter (:py:class:`~onetick.py.core._source._symbol_param._SymbolParamColumn`)
            Pattern to match the value with.
            The pattern can contain usual text characters and two special ones:

            * ``%`` represents zero or more characters
            * ``_`` represents a single character

            Use backslash ``\`` character to escape these special characters.

        Returns
        -------
        Operation
            ``True`` if the match was successful, ``False`` otherwise.
            Note that boolean Operation is converted to float if added as a column.

        Examples
        --------

        Use ``%`` character to specify any number of characters:

        >>> data = otp.Ticks(X=['a', 'ab', 'b_', 'b%'])
        >>> data['LIKE'] = data['X'].str.like('a%')
        >>> otp.run(data)
                             Time   X  LIKE
        0 2003-12-01 00:00:00.000   a   1.0
        1 2003-12-01 00:00:00.001  ab   1.0
        2 2003-12-01 00:00:00.002  b_   0.0
        3 2003-12-01 00:00:00.003  b%   0.0

        Use ``_`` special character to specify a single character:

        >>> data = otp.Ticks(X=['a', 'ab', 'b_', 'b%'])
        >>> data['LIKE'] = data['X'].str.like('a_')
        >>> otp.run(data)
                             Time   X  LIKE
        0 2003-12-01 00:00:00.000   a   0.0
        1 2003-12-01 00:00:00.001  ab   1.0
        2 2003-12-01 00:00:00.002  b_   0.0
        3 2003-12-01 00:00:00.003  b%   0.0

        Use backslash ``\`` character to escape special characters:

        >>> data = otp.Ticks(X=['a', 'ab', 'b_', 'b%'])
        >>> data['LIKE'] = data['X'].str.like(r'b\_')
        >>> otp.run(data)
                             Time   X  LIKE
        0 2003-12-01 00:00:00.000   a   0.0
        1 2003-12-01 00:00:00.001  ab   0.0
        2 2003-12-01 00:00:00.002  b_   1.0
        3 2003-12-01 00:00:00.003  b%   0.0

        This function can be used to filter out ticks:

        >>> data = otp.Ticks(X=['a', 'ab', 'b_', 'b%'])
        >>> data, _ = data[data['X'].str.like('a%')]
        >>> otp.run(data)
                             Time   X
        0 2003-12-01 00:00:00.000   a
        1 2003-12-01 00:00:00.001  ab

        ``pattern`` can only be a constant expression, like string or symbol parameter:

        >>> data = otp.Ticks(X=['a', 'ab', 'b_', 'b%'])
        >>> data['LIKE'] = data['X'].str.like(data.Symbol['PATTERN', str])
        >>> otp.run(data, symbols=otp.Tick(SYMBOL_NAME='COMMON::AAPL', PATTERN='_'))['COMMON::AAPL']
                             Time   X  LIKE
        0 2003-12-01 00:00:00.000   a   1.0
        1 2003-12-01 00:00:00.001  ab   0.0
        2 2003-12-01 00:00:00.002  b_   0.0
        3 2003-12-01 00:00:00.003  b%   0.0
        """
        from onetick.py.core._source._symbol_param import _SymbolParamColumn
        if not isinstance(pattern, (str, _SymbolParamColumn)):
            raise ValueError('like() function expects parameter to be a constant expression')
        return _StrAccessor.Formatter(
            op_params=[self._base_column, pattern],
            dtype=bool,
            formatter=lambda column, pattern: f'{ott.value2str(column)} LIKE {ott.value2str(pattern)}'
        )

    def ilike(self, pattern):
        r"""
        Check if the value is case insensitive matched with SQL-like ``pattern``.

        Parameters
        ----------
        pattern: str or symbol parameter (:py:class:`~onetick.py.core._source._symbol_param._SymbolParamColumn`)
            Pattern to match the value with.
            The pattern can contain usual text characters and two special ones:

            * ``%`` represents zero or more characters
            * ``_`` represents a single character

            Use backslash ``\`` character to escape these special characters.

        Returns
        -------
        Operation
            ``True`` if the match was successful, ``False`` otherwise.
            Note that boolean Operation is converted to float if added as a column.

        Examples
        --------

        Use ``%`` character to specify any number of characters:

        .. testcode::
           :skipif: not is_ilike_supported()

           data = otp.Ticks(X=['a', 'ab', 'Ab', 'b_'])
           data['LIKE'] = data['X'].str.ilike('a%')
           df = otp.run(data)
           print(df)

        .. testoutput::

                                Time   X  LIKE
           0 2003-12-01 00:00:00.000   a   1.0
           1 2003-12-01 00:00:00.001  ab   1.0
           2 2003-12-01 00:00:00.002  Ab   1.0
           3 2003-12-01 00:00:00.003  b_   0.0

        Use ``_`` special character to specify a single character:

        .. testcode::
           :skipif: not is_ilike_supported()

           data = otp.Ticks(X=['a', 'ab', 'Ab', 'b_'])
           data['LIKE'] = data['X'].str.ilike('a_')
           df = otp.run(data)
           print(df)

        .. testoutput::

                                Time   X  LIKE
           0 2003-12-01 00:00:00.000   a   0.0
           1 2003-12-01 00:00:00.001  ab   1.0
           2 2003-12-01 00:00:00.002  Ab   1.0
           3 2003-12-01 00:00:00.003  b_   0.0

        Use backslash ``\`` character to escape special characters:

        .. testcode::
           :skipif: not is_ilike_supported()

           data = otp.Ticks(X=['a', 'ab', 'bb', 'b_'])
           data['LIKE'] = data['X'].str.ilike(r'b\_')
           df = otp.run(data)
           print(df)

        .. testoutput::

                                Time   X  LIKE
           0 2003-12-01 00:00:00.000   a   0.0
           1 2003-12-01 00:00:00.001  ab   0.0
           2 2003-12-01 00:00:00.002  bb   0.0
           3 2003-12-01 00:00:00.003  b_   1.0

        This function can be used to filter out ticks:

        .. testcode::
           :skipif: not is_ilike_supported()

           data = otp.Ticks(X=['a', 'ab', 'Ab', 'b_'])
           data, _ = data[data['X'].str.ilike('a%')]
           df = otp.run(data)
           print(df)

        .. testoutput::

                                Time   X
           0 2003-12-01 00:00:00.000   a
           1 2003-12-01 00:00:00.001  ab
           2 2003-12-01 00:00:00.002  Ab

        ``pattern`` can only be a constant expression, like string or symbol parameter:

        .. testcode::
           :skipif: not is_ilike_supported()

           data = otp.Ticks(X=['a', 'ab', 'A', 'b_'])
           data['LIKE'] = data['X'].str.ilike(data.Symbol['PATTERN', str])
           df = otp.run(data, symbols=otp.Tick(SYMBOL_NAME='COMMON::AAPL', PATTERN='_'))['COMMON::AAPL']
           print(df)

        .. testoutput::

                                Time   X  LIKE
           0 2003-12-01 00:00:00.000   a   1.0
           1 2003-12-01 00:00:00.001  ab   0.0
           2 2003-12-01 00:00:00.002   A   1.0
           3 2003-12-01 00:00:00.003  b_   0.0
        """
        from onetick.py.core._source._symbol_param import _SymbolParamColumn
        if not isinstance(pattern, (str, _SymbolParamColumn)):
            raise ValueError('ilike() function expects parameter to be a constant expression')
        return _StrAccessor.Formatter(
            op_params=[self._base_column, pattern],
            dtype=bool,
            formatter=lambda column, pattern: f'{ott.value2str(column)} ILIKE {ott.value2str(pattern)}'
        )



import typing

from onetick.py import Operation
from onetick.py.core.column_operations.base import _Operation
from onetick.py.types import value2str


class _Accessor:

    def __init__(self, base_column):
        self._base_column = base_column

    class Formatter(_Operation):

        def __init__(self, base_column, dtype, formatter):
            super().__init__(dtype=dtype, op_params=[base_column])
            self._formatter = formatter

        def __str__(self):
            return self._formatter(str(self._op_params[0]))

    def _preprocess_tz_and_format(self,
                                  timezone: typing.Union[Operation, str, None],
                                  format_str):  # it is common for str and dt accessors
        if timezone is None or isinstance(timezone, str) and timezone == "_TIMEZONE":
            timezone = "_TIMEZONE"
        else:
            timezone = value2str(timezone)
        format_str = value2str(format_str)
        return timezone, format_str



from typing import List, Dict, TYPE_CHECKING, Tuple, Optional, Union
from copy import deepcopy
from collections import namedtuple
from onetick.py.compatibility import is_supported_agg_option_price

if TYPE_CHECKING:
    from onetick.py.core.source import Source  # hack for annotations

from onetick.py.core.column import Column
from onetick.py import types as ott
from onetick.py.otq import otq
import onetick.py as otp

from ._base import _Aggregation, operation_gb, operation_replacer, _MultiColumnAggregation
from onetick.py.aggregations.other import (
    FirstTick, LastTick, OptionPrice, Ranking, Percentile, PortfolioPrice, MultiPortfolioPrice,
)
from onetick.py.aggregations.high_low import HighTick, LowTick
from onetick.py.aggregations.generic import Generic


class Compute(_Aggregation):
    @property
    def NAME(self):
        return "COMPUTE"

    @property
    def EP(self):
        try:
            return otq.compute
        except AttributeError:
            return otq.ComputeEp

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['all_fields'] = 'SHOW_ALL_FIELDS'
    FIELDS_TO_SKIP = ['column_name']

    DISALLOWED_FIELDS = ["running", "all_fields", "bucket_interval", "bucket_time",
                         "bucket_units", "end_condition_per_group", "boundary_tick_bucket", "group_by"]
    # [bucket_end_condition, all_fields] can not be used without bucket_units=flexible -> already disallowed

    _validations_to_skip = ['running_all_fields']

    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(column=Column('Time'), *args, **kwargs)

        self.aggrs = {}
        self.has_multi_column_aggregations = False

        # Aggregation from `all_fields` parameter is saved as regular aggregation in `self.aggrs` with empty field name
        if isinstance(self.all_fields, str):
            if self.all_fields.lower() not in ["last", "first", "high", "low", "when_ticks_exit_window"]:
                raise ValueError(f"Unknown all_fields '{self.all_fields}' policy provided.")

            if isinstance(self.all_fields, str) and self.all_fields.lower() != "when_ticks_exit_window":
                if self.running:
                    raise ValueError(
                        "`all_fields` can't be one of 'last', 'first', 'high' or 'low' when `running=True`"
                    )

                aggr_class = globals()[self.all_fields.capitalize() + "Tick"]
                aggr_args = {}
                if self.all_fields in {"high", "low"}:
                    aggr_args['column'] = "PRICE"

                all_fields_aggr = aggr_class(**aggr_args)

                self.add('', all_fields_aggr)

        if isinstance(self.all_fields, (HighTick, LowTick)):
            self.add('', self.all_fields)

        if self.all_fields is True and not self.running:
            self.add('', FirstTick())

    def add(self, name: str, aggr: _Aggregation):
        """Adds aggregation for multiple aggregation"""
        self._validate_aggregation(aggr)
        if name in self.aggrs:
            raise KeyError(f'You are trying to output two aggregations to one field: "{name}"')

        if name == '' and not isinstance(aggr, (FirstTick, LastTick, HighTick, LowTick)):
            raise ValueError(
                'Only one instance of FirstTick, LastTick, HighTick or LowTick aggregation '
                'is allowed to have empty name in order to save compatibility with `all_fields` parameter'
            )

        if aggr.is_multi_column_aggregation:
            self.has_multi_column_aggregations = True

        self.aggrs[name] = aggr

    def _validate_aggregation(self, aggr: _Aggregation):
        if not isinstance(aggr, _Aggregation):
            raise TypeError(f"It is allowed to pass only aggregations, but got '{type(aggr)}'")

        if isinstance(aggr, OptionPrice) and not is_supported_agg_option_price():
            raise NotImplementedError(f".agg() method does not support {type(aggr)} on OneTick {otp.__build__} build")

        if isinstance(aggr, Ranking):
            raise ValueError('Using ranking aggregation in otp.Source.agg method is not supported. '
                             'Use otp.agg.ranking.apply() or otp.Source.ranking() methods.')

        if isinstance(aggr, Percentile):
            raise ValueError('Using percentile aggregation in otp.Source.agg method is not supported. '
                             'Use otp.agg.percentile.apply() or otp.Source.percentile() methods.')

        if isinstance(aggr, PortfolioPrice) and getattr(aggr, 'symbols'):
            raise ValueError('It\'s not allowed to pass `symbols` parameter while using `portfolio_price` '
                             'in compute or `Source.agg`')

        if isinstance(aggr, MultiPortfolioPrice):
            raise ValueError('Using multi_portfolio_price aggregation in otp.Source.agg method is not supported, '
                             'because it outputs multiple ticks at once. '
                             'Use otp.agg.percentile.apply() or otp.Source.percentile() methods.')

        for field in self.DISALLOWED_FIELDS:
            if field not in aggr.FIELDS_TO_SKIP and getattr(aggr, field) != aggr.FIELDS_DEFAULT[field]:
                raise ValueError(f'"{field}" parameter can not be specified in multiple aggregation')

    def validate_input_columns(self, src: 'Source'):
        self._validate_aggregation_with_source(src)

    def _validate_aggregation_with_source(self, src: 'Source'):
        for name, aggr in self.aggrs.items():
            aggr.validate_input_columns(src)

    @property
    def ep_params(self) -> Dict:
        all_fields_orig = self.all_fields
        if self.has_multi_column_aggregations:
            self.all_fields = False
        params = super().ep_params
        if self.has_multi_column_aggregations:
            self.all_fields = all_fields_orig

        params['append_output_field_name'] = self.has_multi_column_aggregations

        what_to_compute = []
        for name, aggr in self.aggrs.items():
            if name:
                str_aggr = str(aggr) + " " + name
            else:
                agg_params = [f'{k}={aggr._attr2str(v)}' for k, v in aggr.ep_params.items()]
                agg_params.append("KEEP_INITIAL_SCHEMA='true'")
                str_aggr = aggr.NAME + "(" + ",".join(agg_params) + ")"

            what_to_compute.append(str_aggr)

        if what_to_compute:
            params['COMPUTE'] = ','.join(what_to_compute)

        return params

    def to_ep(self, *args, **kwargs) -> otq.EpBase:
        params = dict((k.lower(), v) for k, v in self.ep_params.items())
        return self.EP(**params)

    @operation_gb
    @operation_replacer
    def apply(self, src: 'Source') -> 'Source':
        if not self.aggrs:
            raise TypeError('Nothing to aggregate')

        res = src.copy()
        res, what_to_wrap = self._ts_to_long(res)
        schema = self._get_common_schema(res, [key for key in self.aggrs.keys() if key])

        # run _modify_source like it is in apply
        for aggr in self.aggrs.values():
            if isinstance(aggr, Generic):
                aggr_schema = aggr._detect_query_fun_schema(res)

                if aggr._query is None:
                    raise RuntimeError('Generic aggregation was not initialized correctly.')

                # keep only fields from schema
                query = aggr._query.table(**aggr_schema, strict=True)

                # add LAST_TICK EP
                default_tick = ','.join(f'{field} {ott.type2str(dtype)}' for field, dtype in aggr_schema.items())
                aggr._query = query.sink(otq.LastTick(default_tick=default_tick), inplace=False)

                # set empty query params for Generic aggregations
                aggr._set_query_params()

            aggr._modify_source(res)

        # it's important to validate input schema before sinking
        res.sink(self.to_ep())

        for name, aggr in self.aggrs.items():
            if not name:
                # No need to check `all_fields` aggregation: it will have the same schema, as input source
                continue

            if not aggr.is_multi_column_aggregation:
                agg_out_schema = aggr._get_output_schema(res, name=name)
            else:
                aggr_schema_base = aggr._get_output_schema(res)
                if aggr.is_all_columns_aggregation:
                    aggr_schema_base = src.schema

                agg_out_schema = {f'{name}.{key}': value for key, value in aggr_schema_base.items()}

            schema.update(agg_out_schema)

        res.schema.set(**schema)

        if self.has_multi_column_aggregations:
            res = self._remove_all_fields_tick_suffix(res)

        res = self._long_to_ts(res, what_to_wrap)
        return res

    def _ts_to_long(self, src: 'Source') -> Tuple['Source', List]:
        """Convert nsectime columns to long, so aggregation can work in a proper way"""
        res = src.copy()
        forward_result = namedtuple('forward_result', ('aggr', 'columns'))
        what_to_wrap = []
        for name, aggr in self.aggrs.items():
            if '_ts_to_long' in dir(aggr) and callable(aggr._ts_to_long):
                res, col, convert_back = aggr._ts_to_long(res, name)
                if convert_back:
                    what_to_wrap.append(forward_result(aggr, col))
        for r in what_to_wrap:
            self.aggrs[r.columns.tmp_out_column] = self.aggrs.pop(r.columns.out_column)
        return res, what_to_wrap

    def _long_to_ts(self, src: 'Source', what_to_wrap: List) -> 'Source':
        """Convert long columns to nsectime if needed (used along with _ts_to_long)"""
        res = src.copy()
        for r in what_to_wrap:
            res = r.aggr._long_to_ts(res, r.columns)
            self.aggrs[r.columns.out_column] = self.aggrs.pop(r.columns.tmp_out_column)
        return res

    def _remove_all_fields_tick_suffix(self, src: 'Source') -> 'Source':
        """
        renames columns (in query and schema) after compute with multi-column aggregations
        """
        schema = src.schema.copy()
        rename_rule = {}
        fields_to_drop = []

        for name, aggr in self.aggrs.items():
            if aggr.is_all_columns_aggregation or isinstance(aggr, Generic):
                # In case of Generic: remove TICK_TIME field returned by LAST_TICK from generic aggregation
                field_name = name if name else aggr.DEFAULT_OUTPUT_NAME
                tick_time_field = f'{field_name}.TICK_TIME'
                schema.update(**{tick_time_field: ott.nsectime})
                fields_to_drop.append(tick_time_field)

            if aggr.is_multi_column_aggregation:
                continue

            new_name = f'{name}.{aggr.DEFAULT_OUTPUT_NAME}'
            schema[new_name] = schema.pop(name)
            rename_rule[new_name] = name

        res = src.copy()
        res.schema.set(**schema)

        if fields_to_drop:
            res.drop(fields_to_drop, inplace=True)

        if rename_rule:
            res.rename(rename_rule, inplace=True)

        all_fields_aggr = self.aggrs.get('')
        if all_fields_aggr:
            # drop first_tick columns duplicating group_by columns
            all_fields_tick_name = all_fields_aggr.DEFAULT_OUTPUT_NAME

            for group_by in self.group_by:
                group_by_name = group_by
                if isinstance(group_by, Column):
                    group_by_name = group_by.name
                first_tick_field = f'{all_fields_tick_name}.{group_by_name}'
                res.sink(otq.Passthrough(fields=first_tick_field, drop_fields=True))

            res.sink(otq.RenameFieldsEp(rename_fields=rf"{all_fields_tick_name}\.(.*)=\1", use_regex=True))

        return res



import warnings
from typing import TYPE_CHECKING, Tuple, Union

from onetick import py as otp
from onetick.py import aggregations
from onetick.py.aggregations._docs import (
    _all_fields_with_policy_doc,
    _boundary_tick_bucket_doc,
    _bucket_end_condition_doc,
    _bucket_interval_doc,
    _bucket_time_doc,
    _bucket_units_doc,
    _end_condition_per_group_doc,
    _group_by_doc,
    _groups_to_display_doc,
    _running_doc,
    copy_method,
)
from onetick.py.docs.utils import docstring, param_doc
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


_agg_doc = param_doc(
    name='aggs',
    annotation=dict,
    str_annotation='dict of aggregations',
    desc="""
    aggregation dict:

    * key - output column name for regular aggregations, prefix for column names for tick and multi-column aggregations;
    * value - aggregation
    """,
)


@docstring(
    parameters=[
        _agg_doc,
        _running_doc,
        _all_fields_with_policy_doc,
        _bucket_interval_doc,
        _bucket_time_doc,
        _bucket_units_doc,
        _bucket_end_condition_doc,
        _end_condition_per_group_doc,
        _boundary_tick_bucket_doc,
        _group_by_doc,
        _groups_to_display_doc,
    ],
    add_self=True,
)
def agg(self: 'Source', aggs, *args, **kwargs) -> 'Source':
    """
    Applies composition of :ref:`otp.agg <aggregations_funcs>` aggregations

    See Also
    --------
    | :ref:`Aggregations <aggregations_funcs>`
    | **COMPUTE** OneTick event processor

    Returns
    -------
        :py:class:`~onetick.py.Source`

    Examples
    --------

    By default the whole data is aggregated:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'X_SUM': otp.agg.sum('X')})
    >>> otp.run(data)
            Time  X_SUM
    0 2003-12-04     10

    Multiple aggregations can be applied at the same time:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'X_SUM': otp.agg.sum('X'),
    ...                  'X_MEAN': otp.agg.average('X')})
    >>> otp.run(data)
            Time  X_SUM  X_MEAN
    0 2003-12-04     10     2.5

    Aggregation can be used in running mode:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'CUM_SUM': otp.agg.sum('X')}, running=True)
    >>> otp.run(data)
                         Time  CUM_SUM
    0 2003-12-01 00:00:00.000        1
    1 2003-12-01 00:00:00.001        3
    2 2003-12-01 00:00:00.002        6
    3 2003-12-01 00:00:00.003       10

    Aggregation can be split in buckets:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.agg({'X_SUM': otp.agg.sum('X')}, bucket_interval=2, bucket_units='ticks')
    >>> otp.run(data)
                         Time  X_SUM
    0 2003-12-01 00:00:00.001      3
    1 2003-12-01 00:00:00.003      7

    Running aggregation can be used with buckets too. In this case (all_fields=False and running=True) output ticks
    are created when a tick enters or leaves the sliding window (that's why for this example there are 8 output
    ticks for 4 input ticks):

    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3600])
    >>> data = data.agg(dict(X_MEAN=otp.agg.average("X"),
    ...                      X_STD=otp.agg.stddev("X")),
    ...                 running=True, bucket_interval=2)
    >>> otp.run(data)
                         Time  X_MEAN     X_STD
    0 2003-12-01 00:00:00.000     1.0  0.000000
    1 2003-12-01 00:00:01.000     1.5  0.500000
    2 2003-12-01 00:00:01.500     2.0  0.816497
    3 2003-12-01 00:00:02.000     2.5  0.500000
    4 2003-12-01 00:00:03.000     3.0  0.000000
    5 2003-12-01 00:00:03.500     NaN       NaN
    6 2003-12-01 00:00:03.600     4.0  0.000000
    7 2003-12-01 00:00:05.600     NaN       NaN

    By default, if you run aggregation with buckets and group_by, then a bucket will be taken first, and after that
    grouping and aggregation will be performed:

    >>> ticks = otp.Ticks(
    ...     {
    ...         'QTY': [10, 2, 30, 4, 50],
    ...         'TRADER': ['A', 'B', 'A', 'B', 'A']
    ...     }
    ... )
    >>>
    >>> ticks = ticks.agg(
    ...     {'SUM_QTY': otp.agg.sum('QTY')}, group_by='TRADER',
    ...     bucket_interval=3, bucket_units='ticks',
    ...     running=True, all_fields=True,
    ... )
    >>>
    >>> otp.run(ticks)
                         Time  TRADER  QTY  SUM_QTY
    0 2003-12-01 00:00:00.000       A   10       10
    1 2003-12-01 00:00:00.001       B    2        2
    2 2003-12-01 00:00:00.002       A   30       40
    3 2003-12-01 00:00:00.003       B    4        6
    4 2003-12-01 00:00:00.004       A   50       80

    In the row with index 4, the result of summing up the trades for trader "A" turned out to be 80, instead of 90.
    We first took a bucket of 3 ticks, then within it took the group with trader "A" (2 ticks remained) and
    added up the volumes.
    To prevent this behaviour, and group ticks first, set parameter ``end_condition_per_group`` to True:

    >>> ticks = otp.Ticks(
    ...     {
    ...         'QTY': [10, 2, 30, 4, 50],
    ...         'TRADER': ['A', 'B', 'A', 'B', 'A']
    ...     }
    ... )
    >>>
    >>> ticks = ticks.agg(
    ...     {'SUM_QTY': otp.agg.sum('QTY')}, group_by='TRADER',
    ...     bucket_interval=3, bucket_units='ticks',
    ...     running=True, all_fields=True,
    ...     end_condition_per_group=True,
    ... )
    >>>
    >>> otp.run(ticks)
                         Time  TRADER  QTY  SUM_QTY
    0 2003-12-01 00:00:00.000       A   10       10
    1 2003-12-01 00:00:00.001       B    2        2
    2 2003-12-01 00:00:00.002       A   30       40
    3 2003-12-01 00:00:00.003       B    4        6
    4 2003-12-01 00:00:00.004       A   50       90

    Tick aggregations and aggregations, which return more than one output column, could be also used.
    Dict key set for an aggregation in ``aggs`` parameter will be used as prefix
    for each output column of this aggregation.

    >>> data = otp.Ticks(X=[1, 2, 3, 4], Y=[10, 20, 30, 40])
    >>> data = data.agg({'X_SUM': otp.agg.sum('X'), 'X_FIRST': otp.agg.first_tick()})
    >>> otp.run(data)
            Time  X_FIRST.X  X_FIRST.Y  X_SUM
    0 2003-12-04          1         10     10

    These aggregations can be split in buckets too:

    >>> data = otp.Ticks(X=[1, 2, 3, 4], Y=[10, 20, 30, 40])
    >>> data = data.agg(
    ...     {'X_SUM': otp.agg.sum('X'), 'X_FIRST': otp.agg.first_tick()},
    ...     bucket_interval=2, bucket_units='ticks',
    ... )
    >>> otp.run(data)
                         Time  X_FIRST.X  X_FIRST.Y  X_SUM
    0 2003-12-01 00:00:00.001          1         10      3
    1 2003-12-01 00:00:00.003          3         30      7

    If all_fields=True an output tick is generated only for arrival events, but all attributes from the input tick
    causing an arrival event are copied over to the output tick and the aggregation is added as another attribute:

    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3600])
    >>> data = data.agg(dict(X_MEAN=otp.agg.average("X"),
    ...                      X_STD=otp.agg.stddev("X")),
    ...                 all_fields=True, running=True)
    >>> otp.run(data)
                         Time  X  X_MEAN     X_STD
    0 2003-12-01 00:00:00.000  1     1.0  0.000000
    1 2003-12-01 00:00:01.000  2     1.5  0.500000
    2 2003-12-01 00:00:01.500  3     2.0  0.816497
    3 2003-12-01 00:00:03.600  4     2.5  1.118034

    ``all_fields`` parameter can be used when there is need to have all original fields in the output:

    >>> ticks = otp.Ticks(X=[3, 4, 1, 2])
    >>> data = ticks.agg(dict(X_MEAN=otp.agg.average("X"),
    ...                       X_STD=otp.agg.stddev("X")),
    ...                  all_fields=True)
    >>> otp.run(data)
            Time  X  X_MEAN     X_STD
    0 2003-12-04  3     2.5  1.118034

    There are different politics for ``all_fields`` parameter:

    >>> data = ticks.agg(dict(X_MEAN=otp.agg.average("X"),
    ...                       X_STD=otp.agg.stddev("X")),
    ...                  all_fields="last")
    >>> otp.run(data)
            Time  X  X_MEAN     X_STD
    0 2003-12-04  2     2.5  1.118034

    For low/high policies the field selected as input is set this way:

    >>> data = ticks.agg(dict(X_MEAN=otp.agg.average("X"),
    ...                       X_STD=otp.agg.stddev("X")),
    ...                  all_fields=otp.agg.low_tick(data["X"]))
    >>> otp.run(data)
            Time  X  X_MEAN     X_STD
    0 2003-12-04  1     2.5  1.118034

    Example of using 'flexible' buckets. Here every bucket consists of consecutive upticks.

    >>> trades = otp.Ticks(PRICE=[194.65, 194.65, 194.65, 194.75, 194.75, 194.51, 194.70, 194.71, 194.75, 194.71])
    >>> trades = trades.agg({'COUNT': otp.agg.count(),
    ...                     'FIRST_TIME': otp.agg.first('Time'),
    ...                     'LAST_TIME': otp.agg.last('Time')},
    ...                     bucket_units='flexible',
    ...                     bucket_end_condition=trades['PRICE'] < trades['PRICE'][-1])
    >>> otp.run(trades)
                         Time  COUNT              FIRST_TIME               LAST_TIME
    0 2003-12-01 00:00:00.005      5 2003-12-01 00:00:00.000 2003-12-01 00:00:00.004
    1 2003-12-01 00:00:00.009      4 2003-12-01 00:00:00.005 2003-12-01 00:00:00.008
    2 2003-12-04 00:00:00.000      1 2003-12-01 00:00:00.009 2003-12-01 00:00:00.009
    """

    aggs = aggs.copy()
    result = self.copy()

    what_to_aggregate = aggregations.compute(
        *args,
        **kwargs,
    )

    for name, ag in aggs.items():
        what_to_aggregate.add(name, ag)

    result = what_to_aggregate.apply(result)
    result._add_table()

    return result


# Aggregations copy
# we need this functions to store and collect documentation
# copy_method decorator will
#       set docstring (will compare docstring of donor function and method docstring)
#       apply same signature from donor function + self
#       for mimic=True will apply agg function as is
@copy_method(aggregations.functions.high_tick)
def high(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> data = data.high(['X'], 2)     # OTdirective: snippet-name: Aggregations.high tick;
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:01.500  3
    1 2003-12-01 00:00:03.000  4
    """
    pass


@copy_method(aggregations.functions.low_tick)
def low(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[0, 1000, 1500, 3000])
    >>> data = data.low(['X'],2)      # OTdirective: snippet-name: Aggregations.low tick;
    >>> otp.run(data)
                     Time  X
    0 2003-12-01 00:00:00  1
    1 2003-12-01 00:00:01  2
    """
    pass


@copy_method(aggregations.functions.first_tick)
def first(self, *args, **kwargs):
    """
    Examples
    --------

    Get first tick:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.first()    # OTdirective: snippet-name: Aggregations.first;
    >>> otp.run(data)
            Time  X
    0 2003-12-01  1

    Get first tick each day:

    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[otp.Day(0), otp.Day(0), otp.Day(2), otp.Day(2)])
    >>> data = data.first(bucket_interval=1, bucket_units='days', bucket_time='start')
    >>> otp.run(data)
            Time  X
    0 2003-12-01  1
    1 2003-12-03  3

    Get first tick each day and set tick value for empty buckets:

    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[otp.Day(0), otp.Day(0), otp.Day(2), otp.Day(2)])
    >>> data = data.first(bucket_interval=1, bucket_units='days', bucket_time='start', default_tick={'X': -1})
    >>> otp.run(data)
            Time   X
    0 2003-12-01   1
    1 2003-12-02  -1
    2 2003-12-03   3
    """
    pass


@copy_method(aggregations.functions.last_tick)
def last(self, *args, **kwargs):
    """
    Examples
    --------

    Get last tick:

    >>> data = otp.Ticks(X=[1, 2, 3, 4])
    >>> data = data.last()     # OTdirective: snippet-name: Aggregations.last;
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.003  4

    Get last tick each day:

    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[otp.Day(0), otp.Day(0), otp.Day(2), otp.Day(2)])
    >>> data = data.last(bucket_interval=1, bucket_units='days', bucket_time='start')
    >>> otp.run(data)
            Time  X
    0 2003-12-01  2
    1 2003-12-03  4

    Get last tick each day and set tick value for empty buckets:

    >>> data = otp.Ticks(X=[1, 2, 3, 4], offset=[otp.Day(0), otp.Day(0), otp.Day(2), otp.Day(2)])
    >>> data = data.last(bucket_interval=1, bucket_units='days', bucket_time='start', default_tick={'X': -1})
    >>> otp.run(data)
            Time   X
    0 2003-12-01   2
    1 2003-12-02  -1
    2 2003-12-03   4
    """
    pass


@copy_method(aggregations.functions.distinct, mimic=False)
def distinct(self: 'Source', *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.Ticks(dict(x=[1, 3, 1, 5, 3]))
    >>> data = data.distinct('x')   # OTdirective: snippet-name: Aggregations.distinct;
    >>> otp.run(data)
            Time  x
    0 2003-12-04  1
    1 2003-12-04  3
    2 2003-12-04  5
    """
    if 'bucket_interval_units' in kwargs:
        kwargs['bucket_units'] = kwargs.pop('bucket_interval_units')
    aggr = aggregations.functions.distinct(*args, **kwargs)
    return aggr.apply(self)


# mimic=False for backward compatibility
@copy_method(aggregations.functions.high_time, mimic=False, drop_examples=True)
def high_time(self: 'Source', *args, **kwargs):
    """
    Returns timestamp of tick with the highest value of input field

        .. deprecated:: 1.14.5

        Use :py:func:`.high_time` instead

    See Also
    --------
    :py:func:`.high_time`

    """
    warnings.warn(
        f"{self.__class__.__name__}.high_time deprecated. Use otp.agg.high_time instead",
        FutureWarning,
        stacklevel=2,
    )
    aggr = aggregations.functions.high_time(*args, **kwargs)
    return aggr.apply(self, 'VALUE')


# mimic=False for backward compatibility
@copy_method(aggregations.functions.low_time, mimic=False, drop_examples=True)
def low_time(self: 'Source', *args, **kwargs):
    """
    Returns timestamp of tick with the lowest value of input field

        .. deprecated:: 1.14.5

        Use :py:func:`.low_time` instead

    See Also
    --------
    :py:func:`.low_time`

    """
    warnings.warn(
        f"{self.__class__.__name__}.low_time deprecated. Use otp.agg.low_time instead",
        FutureWarning,
        stacklevel=2,
    )
    aggr = aggregations.functions.low_time(*args, **kwargs)
    return aggr.apply(self, 'VALUE')


@copy_method(aggregations.functions.ob_snapshot)
def ob_snapshot(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = data.ob_snapshot(max_levels=1)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  PRICE             UPDATE_TIME  SIZE  LEVEL  BUY_SELL_FLAG
    0 2003-12-04    2.0 2003-12-01 00:00:00.003     6      1              1
    1 2003-12-04    5.0 2003-12-01 00:00:00.004     7      1              0
    """
    pass


@copy_method(aggregations.functions.ob_snapshot_wide)
def ob_snapshot_wide(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = data.ob_snapshot_wide(max_levels=1)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE         BID_UPDATE_TIME  BID_SIZE  ASK_PRICE         ASK_UPDATE_TIME  ASK_SIZE  LEVEL
    0 2003-12-03        5.0 2003-12-01 00:00:00.004         7        2.0 2003-12-01 00:00:00.003         6      1
    """
    pass


@copy_method(aggregations.functions.ob_snapshot_flat)
def ob_snapshot_flat(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = data.ob_snapshot_flat(max_levels=1)  # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE1        BID_UPDATE_TIME1  BID_SIZE1  ASK_PRICE1        ASK_UPDATE_TIME1  ASK_SIZE1
    0 2003-12-03         5.0 2003-12-01 00:00:00.004          7         2.0 2003-12-01 00:00:00.003          6
    """
    pass


@copy_method(aggregations.functions.ob_summary)
def ob_summary(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = data.ob_summary(max_levels=1) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  BID_PRICE  BID_SIZE  BID_VWAP  BEST_BID_PRICE  WORST_BID_SIZE  NUM_BID_LEVELS  ASK_SIZE\
                ASK_VWAP  BEST_ASK_PRICE  WORST_ASK_PRICE  NUM_ASK_LEVELS
    0 2003-12-04        NaN         7       5.0             5.0             NaN               1         6\
            2.0             2.0              2.0               1
    """
    pass


@copy_method(aggregations.functions.ob_size)
def ob_size(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = data.ob_size(max_levels=10) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  ASK_VALUE  BID_VALUE
    0 2003-12-01      84800      64500
    """
    pass


@copy_method(aggregations.functions.ob_vwap)
def ob_vwap(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = data.ob_vwap(max_levels=10) # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  ASK_VALUE  BID_VALUE
    0 2003-12-01     23.313   23.20848
    """
    pass


@copy_method(aggregations.functions.ob_num_levels)
def ob_num_levels(self, *args, **kwargs):
    """
    Examples
    --------
    >>> data = otp.DataSource(db='SOME_DB', tick_type='PRL', symbols='AA')  # doctest: +SKIP
    >>> data = data.ob_num_levels() # doctest: +SKIP
    >>> otp.run(data) # doctest: +SKIP
            Time  ASK_VALUE  BID_VALUE
    0 2003-12-01        248         67
    """
    pass


@copy_method(aggregations.functions.ranking)
def ranking(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.percentile)
def percentile(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.find_value_for_percentile)
def find_value_for_percentile(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.exp_w_average)
def exp_w_average(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.exp_tw_average)
def exp_tw_average(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.standardized_moment)
def standardized_moment(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.portfolio_price)
def portfolio_price(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.multi_portfolio_price)
def multi_portfolio_price(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.return_ep)
def return_ep(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.implied_vol)
def implied_vol(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@copy_method(aggregations.functions.linear_regression)
def linear_regression(self: 'Source', *args, **kwargs):
    # method implementation is copied by decorator
    pass


@inplace_operation
def process_by_group(
    self: 'Source', process_source_func, group_by=None, source_name=None, num_threads=None, inplace=False
) -> Union['Source', Tuple['Source', ...], None]:
    """
    Groups data by ``group_by`` and run ``process_source_func`` for each group and merge outputs for every group.
    Note ``process_source_func`` will be converted to Onetick object and passed to query,
    that means that python callable will be called only once.

    Parameters
    ----------
    process_source_func: callable
        ``process_source_func`` should take :class:`Source` apply necessary logic and return it
        or tuple of :class:`Source` in this case all of them should have a common root that is the
        input :class:`Source`.
        The number of sources returned by this method is the same as the number of sources
        returned by ``process_source_func``.
    group_by: list
        A list of field names to group input ticks by.

        If group_by is None then no group_by fields are defined
        and logic of ``process_source_func`` is applied to all input ticks
        at once
    source_name: str
        A name for the source that represents all of group_by sources. Can be passed here or as a name
        of the inner sources; if passed by both ways, should be consistent
    num_threads: int
        If specified and not zero, turns on asynchronous processing mode
        and specifies number of threads to be used for processing input ticks.
        If this parameter is not specified or zero, then input ticks are processed synchronously.
    inplace: bool
        If True - nothing will be returned and changes will be applied to current query
        otherwise changes query will be returned.
        Error is raised if ``inplace`` is set to True
        and multiple sources returned by ``process_source_func``.

    Returns
    -------
    :class:`Source`, Tuple[:class:`Source`] or None:

    See also
    --------
    **GROUP_BY** OneTick event processor

    Examples
    --------

    >>> # OTdirective: snippet-name: Arrange.group.single output;
    >>> d = otp.Ticks(X=[1, 1, 2, 2],
    ...               Y=[1, 2, 3, 4])
    >>>
    >>> def func(source):
    ...     return source.first()
    >>>
    >>> res = d.process_by_group(func, group_by=['X'])
    >>> otp.run(res)[["X", "Y"]]
       X  Y
    0  1  1
    1  2  3

    Set asynchronous processing:

    >>> res = d.process_by_group(func, group_by=['X'], num_threads=2)
    >>> otp.run(res)[['X', 'Y']]
       X  Y
    0  1  1
    1  2  3

    Return multiple outputs, each with unique grouping logic:

    >>> d = otp.Ticks(X=[1, 1, 2, 2],
    ...               Y=[1, 2, 1, 3])
    >>>
    >>> def func(source):
    ...     source['Z'] = source['X']
    ...     source2 = source.copy()
    ...     source = source.first()
    ...     source2 = source2.last()
    ...     return source, source2
    >>> # OTdirective: snippet-name: Arrange.group.multiple output;
    >>> res1, res2 = d.process_by_group(func, group_by=['Y'])
    >>> df1 = otp.run(res1)
    >>> df2 = otp.run(res2)
    >>> df1[['X', 'Y', 'Z']]
       X  Y  Z
    0  1  1  1
    1  1  2  1
    2  2  3  2
    >>> df2[['X', 'Y', 'Z']]    # OTdirective: skip-snippet:;
       X  Y  Z
    0  1  2  1
    1  2  1  2
    2  2  3  2
    """

    if group_by is None:
        group_by = []

    if inplace:
        main_source = self
    else:
        main_source = self.copy()

    input_schema = main_source.columns(skip_meta_fields=True)
    for field in group_by:
        if field not in input_schema:
            raise ValueError(f"Group by field name {field} not present in input source schema")

    process_source_root = otp.DataSource(tick_type="ANY", schema_policy="manual", schema=input_schema)
    if source_name:
        process_source_root.set_name(source_name)
    process_sources = process_source_func(process_source_root)

    if isinstance(process_sources, otp.Source):
        # returned one source
        process_sources = [process_sources]
    elif len(process_sources) == 1:
        # returned one source as an iterable
        pass
    else:
        # returned multiple sources
        if inplace:
            raise ValueError("Cannot use inplace=True with multi-source processing function!")

    num_source = 0
    for process_source in process_sources:
        output_schema = process_source.columns(skip_meta_fields=True)

        if process_source.get_name():
            if not process_source_root.get_name():
                process_source_root.set_name(process_source.get_name())
            if process_source_root.get_name() != process_source.get_name():
                warnings.warn(
                    "Different strings passed as names for the root source used in "
                    f"process_by_group: '{process_source.get_name()}' "
                    f"and '{process_source_root.get_name()}'"
                )

        # removing key fields from output schema since they will be
        # added by the GROUP_BY EP
        process_source.drop([field for field in group_by if field in output_schema], inplace=True)
        process_source.sink(otq.Passthrough().node_name(f"OUT_{num_source}"))
        process_source_root.node().add_rules(process_source.node().copy_rules())
        main_source._merge_tmp_otq(process_source)
        num_source += 1

    query_name = process_source_root._store_in_tmp_otq(
        main_source._tmp_otq, operation_suffix="group_by", add_passthrough=False,
        # set default symbol, even if it's not set by user, symbol's value doesn't matter in this case
        symbols=otp.config.get('default_symbol', 'ANY')
    )
    process_path = f'THIS::{query_name}'
    num_outputs = len(process_sources)

    # we shouldn't set named outputs if GROUP_BY EP has only one output due to onetick behaviour
    if num_outputs == 1:
        outputs = ""
    else:
        outputs = ",".join([f"OUT_{i}" for i in range(0, num_outputs)])

    kwargs = {}
    if num_threads is not None:
        if num_threads < 0:
            raise ValueError("Parameter 'num_threads' can't be negative.")
        kwargs['num_threads'] = num_threads

    main_source.sink(otq.GroupBy(key_fields=",".join(group_by), query_name=process_path, outputs=outputs, **kwargs))

    output_sources = []
    for num_output in range(0, num_outputs):
        if num_outputs == 1 and inplace:
            output_source = main_source
        else:
            output_source = main_source.copy()

        if num_outputs > 1:
            output_source.node().out_pin(f"OUT_{num_output}")

        # setting schema after processing
        output_schema = process_sources[num_output].columns(skip_meta_fields=True)
        for field in group_by:
            output_schema[field] = input_schema[field]
        for field, field_type in output_schema.items():
            output_source.schema[field] = field_type
        output_source = output_source[list(output_schema)]
        output_source._merge_tmp_otq(main_source)
        output_sources.append(output_source)

    if num_outputs == 1:
        return output_sources[0]
    else:
        return tuple(output_sources)



import doctest
import re
import inspect

from typing import Union, Optional


def snippet_name(example: Union[doctest.Example, str], value):
    if isinstance(example, doctest.Example):
        example.name = value    # type: ignore
    return example


def skip_snippet(example: Union[doctest.Example, str], value):
    if isinstance(example, doctest.Example):
        example.skip_snippet = True    # type: ignore
    return example


def skip_example(example, value, caller=None):
    if caller is not None and value != '' and value != caller:  # do nothing if caller passed and value not caller
        return example
    if isinstance(example, doctest.Example):
        example.skip = True
        return example
    return None


OT_DIRECTIVES = {'skip-example': skip_example,
                 'snippet-name': snippet_name,
                 'skip-snippet': skip_snippet}


def register_ot_directive(directive, fun):
    """
    Register OT directive.

    directive sintax: ...    OTdirective: <directive>: <directive_parameter>;

    Parameters
    ----------
    directive: str
        directive name
    fun: callable
        fun will be executed if `directive` will be found
        fun should have two positional parameters:
            first: string or doctest.Example
            second: direction parameter
        fun should return str, doctest.Example or None
    """
    OT_DIRECTIVES[directive] = fun


class ApplyDirective:

    _OT_DIRECTIVE_RE = r'(.*?)\s*# OTdirective: (.*)'
    _OT_DIRECTIVE_PARSER = re.compile(r'\s*?(?P<func>[A-Za-z_\-0-9]*)\s*?:(?P<param>[^:;]*);')

    def __init__(self, caller: Optional[str] = None):
        self.caller = caller

    def _get_directives(self, item):
        is_example = isinstance(item, doctest.Example)
        string = item
        if is_example:
            string = item.source
        res = []
        funcs = []
        for doc in string.split('\n'):
            s, f = self._get_directives_impl(doc)
            res.append(s)
            funcs.extend(f)

        string = '\n'.join(res)
        if is_example:
            item.source = string
            return item, funcs
        return string, funcs

    def _get_directives_impl(self, string: str):
        directives = re.match(self._OT_DIRECTIVE_RE, string, re.MULTILINE | re.DOTALL)
        if directives is None:
            return string, {}
        doc, directive_str = directives.groups()
        funcs = []
        for fun, v in re.findall(self._OT_DIRECTIVE_PARSER, directive_str):
            if fun.strip() not in OT_DIRECTIVES:
                raise KeyError(f"Unknown directive: '{fun.strip()}'. Original: '{directive_str}'")
            funcs.append((OT_DIRECTIVES[fun.strip()], v.strip()))
        return doc, funcs

    def __call__(self, item):
        doc, funcs = self._get_directives(item)
        for fun, param in funcs:
            params = {}
            if self.caller and 'caller' in inspect.signature(fun).parameters.keys():
                params['caller'] = self.caller
            doc = fun(doc, param, **params)
        return doc


def apply_directive(example: Union[doctest.Example, str], caller: Optional[str] = None):

    """
    Applies OT directive to `example`
    Drops directive comment

    See Also
    --------
    register_ot_directive
    """

    return ApplyDirective(caller)(example)


class OTDoctestParser(doctest.DocTestParser):

    """
    This class applicable only to parse doctest example!
    Do not run tests with this parser!
    """

    def __init__(self, *args, caller: Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.caller = caller

    def _IS_BLANK_OR_COMMENT(self, value):

        if value.strip().startswith('# OTdirective'):
            return False
        else:
            return super()._IS_BLANK_OR_COMMENT(value)

    def parse(self, *args, **kwargs):
        """
        Applies OT directive for all examples
        """
        examples = super().parse(*args, **kwargs)
        res = [apply_directive(x, self.caller) for x in examples]
        return res
